from .algebra_core import AlgebraicAction
from ..expressions.variables import a,b


class difference_of_squares_(AlgebraicAction):
	def __init__(self, **kwargs):
		super().__init__(
			a**2 - b**2,
			(a+b)*(a-b),
			**kwargs
		)



from ..expression_core import *
from .number import Number
from ..combiners.operations import Div
from fractions import Fraction



class Rational(Number, Div):
	value_type = Fraction
	def __init__(self, a, b, **kwargs):
		self.value = Fraction(a, b)
		Div.__init__(self, Smarten(a), Smarten(b))

	def simplify(self):
		pass #idk will make 

	@property
	def numerator(self):
		return self.children[0]

	@property
	def denominator(self):
		return self.children[1]


from ..expressions.combiners.combiners import Combiner
from ..expressions.combiners.scripts import Superscript
from ..expressions.functions.functions import Function, c0, arg
from ..expressions.variables import Variable
from ..expressions.numbers.real import Real
import numpy as np


class Infinity(Real):
    def __init__(self, **kwargs):
        super().__init__(np.inf, '\\infty', 1, **kwargs)
inf = Infinity()

# Or ?
# inf = Real(np.inf, '\\infty', 1)


class Approaches(Combiner):
	symbol = '\\to'
	symbol_glyph_length = 1


class Limit(Function):
	string_code = ['\\lim', '\\limits', '_', c0, arg]
	glyph_code = [3, c0, arg]
	def __init__(self, variable, value, direction='both', **kwargs):
		if direction == 'both':
			destination = value
		elif direction == 'left':
			destination = Superscript(value, Variable('-'))
		elif direction == 'right':
			destination = Superscript(value, Variable('+'))
		else:
			raise ValueError(f"Invalid direction: {direction}. Must be both, left, or right.")
		self.direction = direction
			
		super().__init__(
			children = [Approaches(variable, destination)],
			parentheses_mode = 'weak', 
			**kwargs
		)

	@property
	def variable(self):
		return self.get_subex('00')
	
	@property
	def destination(self):
		destination = self.get_subex('01')
		if isinstance(destination, Superscript):
			return destination.get_subex('0')
		else:
			return destination


from ..expression_core import *
from .combiners import *


class Relation(Combiner):
	def compute(self):
		computed_children = []
		for child in self.children:
			try: child = child.compute()
			except: pass
			computed_children.append(child)
		return all([self.eval_op(computed_children[i], computed_children[i+1]) for i in range(len(self.children)-1)])


class Equation(Relation):
	symbol = '='
	symbol_glyph_length = 1
	eval_op = staticmethod(lambda x, y: x == y)

class LessThan(Relation):
	symbol = '<'
	symbol_glyph_length = 1
	eval_op = staticmethod(lambda x, y: x < y)

class GreaterThan(Relation):
	symbol = '>'
	symbol_glyph_length = 1
	eval_op = staticmethod(lambda x, y: x > y)

class LessThanOrEqualTo(Relation):
	symbol = '\\leq'
	symbol_glyph_length = 1
	eval_op = staticmethod(lambda x, y: x <= y)

class GreaterThanOrEqualTo(Relation):
	symbol = '\\geq'
	symbol_glyph_length = 1
	eval_op = staticmethod(lambda x, y: x >= y)



from .trigonometry_core import TrigFunction
from ..expressions.combiners.operations import Pow


class TrigPower(Pow, TrigFunction):
	def expand_on_args(self, *arg_expressions):
		return self.left(*arg_expressions)**self.right








from ..expression_core import *
from .functions import Function, child, arg
from ..numbers.integer import ten
from ..numbers.real import e
import math


class Log(Function):
	string_code = ['\\log_', child, arg]
	glyph_code = [3, child, arg]
	def __init__(self, base, **kwargs):
		kwargs.setdefault('parentheses_mode', 'strong')
		super().__init__(
			symbol = '\\log',
			children = [base],
			**kwargs
        )

	def python_rule(self, x):
		base = self.base.compute()
		return math.log(x, base)

	@property
	def base(self):
		return self.children[0]



class NaturalLog(Log):
	string_code = ['\\ln', arg]
	glyph_code = [2, arg]
	standard_form = Log(e)
ln = NaturalLog(e)



class CommonLog(Log):
	string_code = ['\\log', arg]
	glyph_code = [3, arg]
	standard_form = Log(ten)
log = CommonLog(10)



from ..expressions.functions import Function
import numpy as np


arcsin = Function('\\arcsin', 6, np.arcsin)
arccos = Function('\\arccos', 6, np.arccos)
arctan = Function('\\arctan', 6, np.arctan)

arcsec = Function('\\arcsec', 6, lambda x: np.arccos(1/x))
arccsc = Function('\\arccsc', 6, lambda x: np.arcsin(1/x))
arccot = Function('\\arccot', 6, lambda x: np.arctan(1/x))


from ..actions import IncompatibleExpression
from ..algebra import AlgebraicAction
from ..timelines import AutoTimeline
from ..expressions import Variables, f, g, e, ln, Number
from .differentials import *
from numpy import pi as PI
TAU = PI*2

x,y,n = Variables('xyn')

class DerivativeRule(AlgebraicAction):
	var_condition_dict = {
		d: lambda exp: isinstance(exp, DifferentialOperator),
		n: lambda exp: isinstance(exp, Number)
	}


class ConstantRule_(DerivativeRule):
	template1 =	d(n)
	template2 =	0

class ConstantMultipleRule_(DerivativeRule):
	template1 =	d(n*x)
	template2 =	n*d(x)

class SumRule_(DerivativeRule):
	template1 =	d(x+y)
	template2 =	d(x) + d(y)
	addressmap = [['+', '+']]

class DifferenceRule_(DerivativeRule):
	template1 =	d(x-y)
	template2 =	d(x) - d(y)
	addressmap = [['-', '-']]

class ProductRule_(DerivativeRule):
	template1 =	d(x*y)
	template2 =	y*d(x) + x*d(y)
	addressmap = [['*', '+'], [[], '0*'], [[], '1*']]

class QuotientRule_(DerivativeRule):
	template1 =	d(x/y)
	template2 =	(y*d(x) - x*d(y)) / y**2

class PowerRule_(DerivativeRule):
	template1 =	d(x**n)
	template2 =	n * x**(n-1) * d(x)
	addressmap = [['11', '0110'], [[], '011-1']]
	var_kwarg_dict = {n:{'path_arc':TAU/3}}

class eRule(DerivativeRule):
	template1 = d(e**x)
	template2 = e**x*d(x)

class ExponentialRule_(DerivativeRule):
	template1 =	d(n**x)
	template2 =	ln(n) * n**x * d(x)
	addressmap = [[[], '000']]

# class ChainRule(DerivativeRule):
	# template1 = d(f(g(a)))
	# template2 = d(f)(g(a))*d(g(a)))
# Idk I think this could be just always built in to all other rules


from ..algebra.simplify import *
Simplify_Rules = [rule() for rule in SimplificationRule.__subclasses__()]
Derivative_Rules = [rule() for rule in DerivativeRule.__subclasses__()]

class Differentiate(AutoTimeline):
	def decide_next_action(self, index):
		last_exp = self.get_expression(-1)
		for ad in last_exp.get_all_addresses():
			for ruleset in [Derivative_Rules, Simplify_Rules]:
				for rule in ruleset:
					try:
						action = rule.copy().pread(ad)
						action.get_output_expression(last_exp)
						return action
					except IncompatibleExpression:
						pass
		for ad in last_exp.get_all_twig_addresses():
			try:
				from ..actions.evaluation import evaluate_
				action = evaluate_().copy().pread(ad)
				action.get_output_expression(last_exp)
				return action
			except IncompatibleExpression:
				pass
		return None





__version__ = '0.5.5'
__author__ = 'John Connell - The Mathematic Fanatic'
__description__ = 'Manim plugin which aims to make it much easier to meaningfully transform algebra expressions.'

from .expressions import *
from .actions import *
from .timelines import *
from .utils import *
from .algebra import *
from .calculus import *
from .trigonometry import *
from .sets import *



from ..expression_core import *
from .number import Number
import numpy as np


class Real(Number):
	decimal_places = algebra_config['decimal_precision']
	internal_precision = 10**-8
	value_type = float
	def __init__(self, value, symbol=None, symbol_glyph_length=None, decimal_places=None, **kwargs):
		super().__init__(**kwargs)
		rounded = round(value, self.decimal_places)
		if np.abs(value - rounded) < self.internal_precision:
			self.value = rounded
		else:
			self.value = value
		self.symbol = symbol
		self.symbol_glyph_length = symbol_glyph_length
		if decimal_places is not None:
			self.decimal_places = decimal_places

	@Expression.parenthesize_glyph_count
	def get_glyph_count(self):
		if self.symbol:
			if self.symbol_glyph_length:
				return self.symbol_glyph_length
		else: # This needs work... parentheses are an issue.
			string = self.__str__.__wrapped__(self) # Ok this might do it but still seems a little stupid
			count = len(string)
			if string.endswith(r'\ldots'): # Like fr? But it works lol
				count -= 3
			return count

	@Expression.parenthesize_latex
	def __str__(self, use_decimal=False):
		if self.symbol and not use_decimal:
			return self.symbol
		rounded = round(self.value, self.decimal_places)
		if rounded == self.value:
			return str(rounded)
		else:
			return f'{self.value:.{self.decimal_places}f}' + r'\ldots'

	def is_negative(self):
		return self.value < 0
	
	def compute(self):
		if self.value.is_integer():
			return int(self.value)
		else:
			return self.value


e = Real(np.e, 'e')
pi = Real(np.pi, '\\pi')
tau = Real(np.pi*2, '\\tau')



from ..expression_core import *
from .combiners import *


class Sequence(Combiner):
	symbol = ','
	symbol_glyph_length = 1
	def __init__(self, *children, generator=None, **kwargs):
		self.generator = generator
		super().__init__(*children, **kwargs)


class Coordinate(Sequence):
	parentheses = True



from .action_core import *
from .animations import *
from .variants import *
from .parallel import *
from .substitution import *
from .permutations import *
from .evaluation import *
from .apply_operation import *



from ..expression_core import *


class Combiner(Expression):
	symbol = None
	symbol_glyph_length = None
	left_spacing = ''
	right_spacing = ''

	@Expression.parenthesize_glyph_count
	def get_glyph_count(self):
		count = sum([child.glyph_count for child in self.children])
		count += self.symbol_glyph_length * (len(self.children) - 1)
		return count

	@Expression.parenthesize_latex
	def __str__(self, *args, **kwargs):
		joiner = self.left_spacing + self.symbol + self.right_spacing
		return joiner.join(['{' + str(child) + '}' for child in self.children])

	def set_spacing(self, left_spacing, right_spacing):
		self.left_spacing = left_spacing
		self.right_spacing = right_spacing

	special_character_to_glyph_method_dict = {
		**Expression.special_character_to_glyph_method_dict,
		'+': 'get_op_glyphs',
		'-': 'get_op_glyphs',
		'*': 'get_op_glyphs',
		'/': 'get_op_glyphs',
		'^': 'get_op_glyphs',
		'=': 'get_op_glyphs',
		'<': 'get_op_glyphs',
		'>': 'get_op_glyphs',
		',': 'get_op_glyphs',
	}

	@Expression.parenthesize_glyph_list
	def get_glyphs_at_addigit(self, addigit:int):
		start = 0
		for sibling in self.children[:addigit]:
			start += sibling.glyph_count
			start += self.symbol_glyph_length
		child = self.children[addigit]
		end = start + child.glyph_count
		return list(range(start, end))

	def get_op_glyphs(self):
		results = []
		turtle = self.parentheses * self.paren_length()
		for child in self.children[:-1]:
			turtle += child.glyph_count
			results += list(range(turtle, turtle + self.symbol_glyph_length))
			turtle += self.symbol_glyph_length
		return results



from .algebra_core import AlgebraicAction
from ..expressions.variables import a
from ..utils import Smarten


class SimplificationRule(AlgebraicAction):
	pass


class add_zero_R(SimplificationRule):
	template1 = a + 0
	template2 = a
	addressmap = [['+1', []]]

class add_zero_L(SimplificationRule):
	template1 = 0 + a
	template2 = a
	addressmap = [['+0', []]]

class sub_zero_R(SimplificationRule):
	template1 = a - 0
	template2 = a
	addressmap = [['-1', []]]

class sub_zero_L(SimplificationRule):
	template1 = 0 - a
	template2 = -a
	addressmap = [['0', []], ['-', '-']]

class mul_zero_R(SimplificationRule):
	template1 = a * 0
	template2 = 0
	addressmap = [['*0', []]]

class mul_zero_L(SimplificationRule):
	template1 = 0 * a
	template2 = 0
	addressmap = [['*1', []]]

class mul_one_L(SimplificationRule):
	template1 = 1 * a
	template2 = a
	addressmap = [['*0', []]]

class mul_one_R(SimplificationRule):
	template1 = a * 1
	template2 = a
	addressmap = [['*1', []]]

class div_one_R(SimplificationRule):
	template1 = a / 1
	template2 = a
	addressmap = [['/1', []]]

class div_zero_L(SimplificationRule):
	template1 = 0 / a
	template2 = 0
	addressmap = [['/1', []]]

class pow_one_R(SimplificationRule):
	template1 = a ** 1
	template2 = a
	addressmap = [['^1', []]]

class pow_one_L(SimplificationRule):
	template1 = 1 ** a
	template2 = 1
	addressmap = [['^1', []]]

class pow_zero_R(SimplificationRule):
	template1 = a ** 0
	template2 = 1
	addressmap = [['', '']]

class pow_zero_L(SimplificationRule):
	template1 = 0 ** a
	template2 = 0
	addressmap = [['^1', []]]



from ..expressions.expression_core import Expression
from ..expressions.combiners.operations import Mul
from .animations import TransformByAddressMap
from MF_Tools.dual_compatibility import Write, FadeIn, FadeOut, TransformMatchingTex
from ..utils import MF_Base, apply_addressmap
from functools import wraps


class Action(MF_Base):
	introducer = Write
	introduce_kwargs = {'run_time':0.5, 'delay':0.5}
	remover = FadeOut
	remove_kwargs = {'run_time':0.5}
	preaddress = ''
	def __init__(self,
		introducer=None,
		remover=None,
		preaddress=''
	):
		self.introducer = introducer or self.introducer
		self.remover = remover or self.remover
		self.preaddress = preaddress or self.preaddress

	def get_output_expression(self, input_expression):
  		# define in subclasses
		return None

	def get_addressmap(self, input_expression, **kwargs):
		# define in subclasses
		return [['', '']]


	### Animating ###

	def get_animation(self, **kwargs):
		def animation(input_exp, output_exp=None, **kwargs):
			if output_exp is None:
				output_exp = self.get_output_expression(input_exp)
			def get_TBAM(action, input_exp, output_exp, **kwargs):
				return TransformByAddressMap(
					input_exp,
					output_exp,
					*action.get_addressmap(input_exp),
					default_introducer=action.introducer,
					default_remover=action.remover,
					**kwargs
				)
			try:
				TBAM = get_TBAM(self.copy(), input_exp.copy(), output_exp.copy())
				assert not TBAM.show_indices, f'Invalid Glyphmap: {TBAM.glyphmap}'
				return get_TBAM(self, input_exp, output_exp)
			except Exception as E:
				print('Warning: Action produced an invalid glyphmap. Falling back to TransformMatchingTex')
				print('Exception: ', E)
				print('Action: ', self)
				print('Input: ', input_exp)
				print('Output: ', output_exp)
				print('Addressmap: ', self.get_addressmap(input_exp))
				return TransformMatchingTex(input_exp.mob, output_exp.mob, **kwargs)
		return animation

	def __call__(self, expr1, expr2=None, **kwargs):
		return self.get_animation(**kwargs)(expr1, expr2)


	### Modifiers ###

	def pread(self, *addresses, **kwargs):
		self = self.copy()
		if len(addresses) == 0:
			return self
		elif len(addresses) == 1:
			self.preaddress = addresses[0] + self.preaddress
			return self
		else:
			from .parallel import ParallelAction
			return ParallelAction(*[self.pread(ad) for ad in addresses], **kwargs)

	def left(self, **kwargs):
		return self.pread('0', **kwargs)

	def right(self, **kwargs):
		return self.pread('1', **kwargs)

	def both(self, number_of_children=2, **kwargs):
		# Intended to turn an action on an expression into an action done to both sides of an equation.
		# Can be passed a number to apply to more than 2 sides for, say, a triple equation or inequality.
		return self.pread(*[str(i) for i in range(number_of_children)], **kwargs)


	### Combinations ###

	def create_parallel(self, other):
		from .parallel import ParallelAction
		if isinstance(other, ParallelAction):
			return ParallelAction(self, *other.actions)
		elif isinstance(other, Action):
			return ParallelAction(self, other)
		else:
			raise ValueError("Can only use |,+ with other ParallelAction or Action")

	def __or__(self, other):
		return self.create_parallel(other)
	
	def __add__(self, other):
		return self.create_parallel(other)

	def __le__(self, expr):
		assert isinstance(expr, Expression), "Can only apply expression >= action"
		return self.get_output_expression(expr)


	### Decorators ###

	def __init_subclass__(cls):
		super().__init_subclass__()
		if not getattr(cls, '_is_decorated', False):
			cls._is_decorated = True

			func = cls.get_output_expression
			func = Action.preaddressfunc(func)
			cls.get_output_expression = func

			func = cls.get_addressmap
			func = Action.preaddressmap(func)
			func = Action.autoparenmap(func)
			# func = Action.autoopmap(func)
			cls.get_addressmap = func


	@staticmethod
	def preaddressfunc(func):
		@wraps(func)
		def wrapper(action, expr, *args, **kwargs):
			expr = expr.copy()
			preaddress = kwargs.get('preaddress', '') or action.preaddress
			active_part = expr.get_subex(preaddress)
			result = func(action, active_part)
			output_expression = expr.substitute_at_address(result, preaddress)
			output_expression.reset_parentheses()
			return output_expression
		return wrapper

	@staticmethod
	def preaddressmap(getmap):
		@wraps(getmap)
		def wrapper(action, expr, *args, **kwargs):
			expr = expr.copy()
			preaddress = kwargs.get('preaddress', '') or action.preaddress
			addressmap = getmap(action, expr, *args, **kwargs)
			if preaddress:
				for entry in addressmap:
					for i, ad in enumerate(entry):
						if isinstance(ad, str):
							entry[i] = preaddress + ad
			return addressmap
		return wrapper

	@staticmethod
	def autoparenmap(getmap, mode='stupid'):
		if mode == 'none':
			return getmap
		if mode == 'stupid':
			@wraps(getmap)
			def wrapper(action, expr, *args, **kwargs):
				addressmap = list(getmap(action, expr, *args, **kwargs))
				in_expr, out_expr = expr, action.get_output_expression(expr)
				for in_add in in_expr.get_all_addresses():
					if in_expr.get_subex(in_add).parentheses:
						addressmap.append([in_add+'()', [], Action.remove_kwargs.copy()])
					for entry in addressmap:
						if entry[0] == in_add:
							entry[0] = entry[0] + '_'
				for out_add in out_expr.get_all_addresses():
					if out_expr.get_subex(out_add).parentheses:
						addressmap.append([[], out_add+'()', Action.introduce_kwargs.copy()])
					for entry in addressmap:
						if entry[1] == out_add:
							entry[1] = entry[1] + '_'
				return addressmap

		if mode == 'smart':
			@wraps(getmap)
			def wrapper(action, expr, *args, **kwargs):
				addressmap = list(getmap(action, expr, *args, **kwargs))
				in_expr, out_expr = expr, action.get_output_expression(expr)
		return wrapper

	@staticmethod
	def autoopmap(getmap, ops_to_check=(Mul,)):

		@wraps(getmap)
		def wrapper(action, expr:Expression, *args, **kwargs):

			addressmap = list(getmap(action, expr, *args, **kwargs))
			in_ads = [entry[0] for entry in addressmap]
			out_ads = [entry[1] for entry in addressmap]

			in_expr, out_expr = expr, action.get_output_expression(expr)
			in_op_ads = in_expr.get_all_addresses_of_type(ops_to_check)
			out_op_ads = out_expr.get_all_addresses_of_type(ops_to_check)

			for ad in in_op_ads:
				
				if any(ad + op in in_ads for op in '+-*/^=<>,'):
					# If op explicitly handled, abort
					continue

				targets = apply_addressmap(ad, addressmap)

				if targets is None or len(targets) != 1:
					# Ambiguous addressmap, just fade out
					addressmap.append( [ad + '+', [], Action.remove_kwargs.copy()] )
				
				# elif len(targets) == 0:
				# 	pass #idk what im doing

				else:
					# Matching address found
					target = targets.pop()
					in_len = in_expr.get_subex(ad).symbol_glyph_length

					if target in out_op_ads:
						# Matching address is also a checked type (??)
						out_op_ads.remove(target)
						out_len = out_expr.get_subex(target).symbol_glyph_length
					else:
						# Idk I'm a bit confused here
						continue

					if in_len and out_len:
						# Both are checked type and have op glyphs, map to each other
						addressmap.append( [ad + '+', target + '+'] )
					
					elif in_len and not out_len:
						# Fade out glyph
						addressmap.append( [ad + '+', [], Action.remove_kwargs.copy()])
					
					elif not in_len and out_len:
						# Fade in glyph
						addressmap.append( [[], target + '+', Action.introduce_kwargs.copy()])
					
					else:
						# Neither has glyph so ignore
						pass
			
			for ad in out_op_ads:

				if any(ad + op in out_ads for op in '+-*/^=<>,'):
					continue

				targets = apply_addressmap(ad, addressmap, reverse=True)

				if targets is None or len(targets) != 1:
					addressmap.append( [[], ad + '+', Action.introduce_kwargs.copy()] )
				
				else:
					raise Exception('This case should have already been caught in the in_op_ads loop.')
			
			return addressmap

		return wrapper





	### Utilities ###

	def __repr__(self):
		max_length = 50
		string = type(self).__name__ + '(' + self.preaddress + ')'
		if len(string) > max_length:
			string = string[:max_length-3] + '...'
		return string



class IncompatibleExpression(Exception):
	pass






from .functions import Function, child, arg, arg0, arg1


class AbsoluteValue(Function):
	string_code = ['\\left|', arg, '\\right|']
	glyph_code = [1, arg, 1]
	def __init__(self, **kwargs):
		super().__init__(
			python_rule = abs,
			parentheses_mode = 'never',
			**kwargs
		)
abs_val = AbsoluteValue()


class Factorial(Function):
	string_code = [arg, '!']
	glyph_code = [arg, 1]
	def __init__(self, **kwargs):
		from scipy.special import gamma
		super().__init__(
			python_rule = lambda z: gamma(z+1),
			parentheses_mode = 'strong',
			**kwargs
		)
fact = Factorial()


from ..variables import n, k
class BinomialCoefficient(Function):
	def __init__(self, mode='binom', **kwargs):
		self.mode = mode
		super().__init__(
			algebra_rule_variables = [n,k],
			algebra_rule = fact(n) / (fact(k) * fact(n-k)),
			parentheses_mode = 'never',
			**kwargs
		)

	@property
	def	string_code(self):
		if self.mode == 'binom':
			return ['\\binom', arg0, arg1]
		elif self.mode == 'nCk':
			return ['_', arg0, 'C', '_', arg1]
		elif self.mode == 'C(n,k)':
			return ['C', arg]
		else:
			raise ValueError(f"Invalid binomial coefficient mode: {self.mode}. Mode must be among ['binom', 'nCk', 'C(n,k)']")

	@property
	def glyph_code(self):
		if self.mode == 'binom':
			return [1, arg0, arg1, 1]
		elif self.mode == 'nCk':
			return [arg0, 1, arg1]
		elif self.mode == 'C(n,k)':
			return Function.glyph_code
		else:
			raise ValueError(f"Invalid binomial coefficient mode: {self.mode}. Mode must be among ['binom', 'nCk', 'C(n,k)']")


binom = BinomialCoefficient()
nCk = BinomialCoefficient(mode='nCk')






from ..actions import IncompatibleExpression
from ..algebra import AlgebraicAction
from ..timelines import AutoTimeline
from ..expressions import Variables, f, g, e, ln, Number
from .integrals import *
from .differentials import DifferentialOperator, d, du, dv, dx
from numpy import pi as PI
TAU = PI*2

a,b,c,n,u,v,x,C = Variables('abcnuvxC')

class IntegralRule(AlgebraicAction):
	var_condition_dict = {
		d: lambda exp: isinstance(exp, DifferentialOperator),
		I: lambda exp: isinstance(exp, IntegralOperator),
		n: lambda exp: isinstance(exp, Number)
	}

class FundThmCalc_1_(IntegralRule):
	template1 = d(DefiniteIntegral(n,x)(u*dx))
	template2 = u

class FundThmCalc_2_(IntegralRule):
	template1 = DefiniteIntegral(a,b)(du)
	template2 = PlugInBounds(a,b,x)(u)

class Int_dx_(IntegralRule):
	template1 = I(d(u))
	template2 = u

class d_Int_x_(IntegralRule):
	template1 = d(I(u))
	template2 = u

class Int_ConstantMultiple_(IntegralRule):
	template1 =	I(n*u)
	template2 =	n*I(u)

class Int_ConstantMultiple_dx_(Int_ConstantMultiple_):
	template1 =	I(n*u*dx)
	template2 =	n*I(u*dx)

class Int_SumRule_(IntegralRule):
	template1 =	I(a+v)
	template2 =	I(a) + I(v)
	addressmap = [['1+', '+']]

class Int_SumRule_dx_(IntegralRule):
	template1 =	I((u+v)*dx)
	template2 =	I(u*dx) + I(v*dx)
	addressmap = [['10+', '+']]

class Int_DifferenceRule_(IntegralRule):
	template1 =	I(a-v)
	template2 =	I(a) - I(v)
	addressmap = [['1-', '-']]

class Int_PowerRule_dx_(IntegralRule):
	template1 = I(x**n*dx)
	template2 = 1/(n+1)*x**(n+1)
	var_condition_dict = {
		n: lambda exp: isinstance(exp, Number) and not exp == -1
	}

class Int_DifferenceSumRule_dx_(IntegralRule):
	template1 =	I((u-v)*dx)
	template2 =	I(u*dx) - I(v*dx)
	addressmap = [['10-', '-']]


class IntegrationByParts_(IntegralRule):
	template1 =	I(u*dv)
	template2 =	u*v - I(v*du)
IBP_indefinite_ = IntegrationByParts_

# class IBP_definite_(IntegralRule):
# 	template1 = DefiniteIntegral(a,b)(u*dv)
# 	template2 = PlugInBounds(a,b)(u*v) - DefiniteIntegral(a,b)(v*du)
# not sure how to do pluginbounds' variable. Do we even need this though?


from ..algebra.simplify import *
Simplify_Rules = [rule() for rule in SimplificationRule.__subclasses__()]
Integral_Rules = [rule() for rule in IntegralRule.__subclasses__()]

class Integrate(AutoTimeline):
	def decide_next_action(self, index):
		last_exp = self.get_expression(-1)
		for ad in last_exp.get_all_addresses():
			for ruleset in [Integral_Rules, Simplify_Rules]:
				for rule in ruleset:
					try:
						action = rule.copy().pread(ad)
						action.get_output_expression(last_exp)
						return action
					except IncompatibleExpression:
						pass
		for ad in last_exp.get_all_twig_addresses():
			try:
				from ..actions.evaluation import evaluate_
				action = evaluate_().copy().pread(ad)
				action.get_output_expression(last_exp)
				return action
			except IncompatibleExpression:
				pass
		return None





from .action_core import Action
from ..expressions.expression_core import Expression
from MF_Tools.dual_compatibility import PI, DOWN, FadeIn, FadeOut
from ..utils import Smarten




class substitute_(Action):
	def __init__(self,
		sub_dict,
		mode = 'transform',
		arc_size = PI,
		fade_shift = DOWN*0.2,
		lag = 0,
		maintain_color = False,
		**kwargs
	):
		self.sub_dict = sub_dict
		self.mode = mode
		self.arc_size = arc_size
		self.fade_shift = fade_shift
		self.lag = lag #usually looks like shit but can be cool sometimes
		self.maintain_color = maintain_color
		super().__init__(**kwargs)

	def get_output_expression(self, input_expression=None):
		result = input_expression.substitute(self.sub_dict)
		if self.maintain_color:
			for from_subex, to_subex in self.sub_dict.items():
				color = input_expression.get_color_of_subex(from_subex)
				addresses = input_expression.get_addresses_of_subex(from_subex)
				for i,ad in enumerate(addresses):
					if input_expression.get_subex(ad).parentheses and not Smarten(to_subex).parentheses:
						addresses[i] += '_'
				result[*addresses].set_color(color)
		return result

	def get_addressmap(self, input_expression=None):
		target_addresses = []
		for var in self.sub_dict:
			target_addresses += input_expression.get_subex(self.preaddress).get_addresses_of_subex(var)
		addressmap = []
		if self.mode == 'transform':
			for i,ad in enumerate(target_addresses):
				addressmap.append([ad, ad, {'delay': self.lag*i}])
		elif self.mode == 'swirl':
			for i,ad in enumerate(target_addresses):
				addressmap.append([ad, ad, {'path_arc': self.arc_size, 'delay': self.lag*i}])
		elif self.mode == 'fade':
			for i,ad in enumerate(target_addresses):
				addressmap.append([ad, FadeOut, {'shift': self.fade_shift, 'delay': self.lag*i}])
				addressmap.append([FadeIn, ad, {'shift': self.fade_shift, 'delay': self.lag*i}])

		# # Horrible bandaid for the problem of multiplication symbols changing and ruining the addressmap
		# # I am certain that there is a general and elegant way to do this, along with parentheses changes,
		# # and other such aesthetic matters which don't change the tree, but I still haven't figured it out.
		# from ..expressions.combiners.operations import Mul
		# input_expression = input_expression.copy()
		# output_expression = self.get_output_expression(input_expression.copy())
		# for ad in output_expression.get_all_addresses():
		# 	in_subex = input_expression.get_subex(ad)
		# 	out_subex = output_expression.get_subex(ad)
		# 	if isinstance(in_subex, Mul) and isinstance(out_subex, Mul):
		# 		if in_subex.symbol == '' and out_subex.symbol != '':
		# 			addressmap.append([ad, ad+'*'])
		# 		if in_subex.symbol != '' and out_subex.symbol == '':
		# 			addressmap.append([ad+'*', ad])

		return addressmap
		
	def __repr__(self):
		return type(self).__name__ + '(' + str(self.sub_dict) + (',' + self.preaddress if self.preaddress else '') + ')'


class substitute_into_(Action):
	def __init__(self, outer_expression, substitution_variable=None, **kwargs):
		self.outer_expression = outer_expression
		if isinstance(substitution_variable, Expression) and substitution_variable.is_variable():
			self.substitution_variable = substitution_variable
		elif substitution_variable is None:
			exp_vars = outer_expression.get_all_variables()
			if len(exp_vars) == 1:
				self.substitution_variable = exp_vars.pop()
			else:
				raise ValueError('Substitution variable must be explicitly given if the expression does not have one variable')
		else:
			raise ValueError(f'Invalid value for substitution_variable: {substitution_variable}')
		super().__init__(**kwargs)

	def get_output_expression(self, input_expression=None):
		return self.outer_expression.substitute({self.substitution_variable: input_expression})

	def get_addressmap(self, input_expression=None):
		sub_into_addresses = self.outer_expression.get_addresses_of_subex(self.substitution_variable)
		addressmap = [
			['', ad]
			for ad in sub_into_addresses
		]
		return addressmap
	
	def get_animation(self, *args, **kwargs):
		return super().get_animation(*args, auto_fade=True, auto_resolve_delay=0.75, **kwargs)



from .set_core import Set
from ..expressions import BinaryOperation, Relation, Pow


class SetRelation(Relation):
	pass


class IsElement(SetRelation):
	symbol = '\\in'
	symbol_glyph_length = 1
	eval_op = staticmethod(lambda a, A: a in A)


class IsSubset(SetRelation):
	symbol = '\\subseteq'
	symbol_glyph_length = 1
	eval_op = staticmethod(lambda A, B: all(a in B for a in A))


class IsProperSubset(SetRelation):
	symbol = '\\subset'
	symbol_glyph_length = 1
	eval_op = staticmethod(lambda A, B: all(a in B for a in A) and any(b in A for b in B))



class SetOperation(Set, BinaryOperation):
	in_op = staticmethod(lambda A, a: a in A)

	def contains(self, item):
		return self.in_op(item)

	def evaluate(self):
		results = set()
		


class Intersection(SetOperation):
	symbol = '\\cap'
	symbol_glyph_length = 1
	in_op = staticmethod(lambda A, a: all(a in Ai for Ai in A.children))


class Union(SetOperation):
	symbol = '\\cup'
	symbol_glyph_length = 1
	in_op = staticmethod(lambda A, a: any(a in Ai for Ai in A.children))


class Difference(SetOperation):
	symbol = '\\setminus'
	symbol_glyph_length = 1
	in_op = staticmethod(lambda A, a: a in A.children[0] and not any(a in Ai for Ai in A.children[1:]))


class CartesianProduct(SetOperation):
	symbol = '\\times'
	symbol_glyph_length = 1
	in_rule = lambda A, a: all(ai in Ai for ai,Ai in zip(a.children,A.children))

	def get_tuple_set(self):
		from .set_core import ElementsSet
		from itertools import product
		return ElementsSet(*product(*self.children))


class CartesianPower(Pow, Set):
	in_rule = lambda An, a: len(a.children) == An.power and all(ai in An.base for ai in a.children)

	@property
	def power(self):
		return self.children[1].compute()

	@property
	def base(self):
		return self.children[0]

	def get_cartesian_product(self):
		return CartesianProduct(*[self.base.copy() for _ in range(self.power)])

	def get_tuple_set(self):
		return self.get_cartesian_product().get_tuple_set()







from ..expressions import *
from ..actions import *
from MF_Tools.dual_compatibility import TransformMatchingTex, UP, smooth, Scene


class Timeline(MF_Base):
	def __init__(
		self,
		auto_color = {},
		auto_scale = 1,
		auto_fit = [None, None, None],
		auto_propagate = True,
		show_past_steps = False,
		past_steps_opacity = 0.4,
		past_steps_direction = UP,
		past_steps_buff = 1,
		past_steps_shift_run_time = 1,
		past_steps_shift_rate_func = smooth
	):
		self.steps = [] # Elements of this list are of the form [expression, action]
		self.current_exp_index = 0
		self.auto_color = auto_color
		self.auto_scale = auto_scale
		self.auto_fit = auto_fit
		self.auto_propagate = auto_propagate
		self.show_past_steps = show_past_steps
		if self.show_past_steps:
			self.past_steps_vgroup = VGroup()
			self.past_steps_opacity = past_steps_opacity
			self.past_steps_direction = past_steps_direction
			self.past_steps_buff = past_steps_buff
			self.past_steps_shift_run_time = past_steps_shift_run_time
			self.past_steps_shift_rate_func = past_steps_shift_rate_func

	def get_expression(self, index: int) -> Expression:
		try:
			return self.steps[index][0]
		except IndexError:
			return None

	def set_expression(self, index: int, expression: Expression):
		if self.auto_color:
			expression.set_color_by_subex(self.auto_color)
		if any(self.auto_fit):
			expression.mob.scale_to_fit(*self.auto_fit)
		elif self.auto_scale != 1:
			expression.mob.scale(self.auto_scale)
		if index == len(self.steps):
			self.add_expression_to_end(expression)
		self.steps[index][0] = expression
		if self.auto_propagate:
			self.propagate(start_at=index)

	def add_expression_to_start(self, expression: Expression):
		if len(self.steps) == 0 or self.steps[0][0] is not None:
			self.steps.insert(0, [None, None])
		self.set_expression(0, expression)
		return self

	def add_expression_to_end(self, expression: Expression):
		self.steps.append([None, None])
		self.set_expression(-1, expression)
		return self

	def get_action(self, index: int) -> Action:
		try:
			return self.steps[index][1]
		except IndexError:
			return None

	def set_action(self, index: int, action: Action):
		self.steps[index][1] = action
		if self.auto_propagate:
			self.propagate(start_at=index)

	def add_action_to_start(self, action: Action):
		self.set_action(0, action)
		return self

	def add_action_to_end(self, action: Action):
		if len(self.steps) == 0 or self.steps[-1][1] is not None:
			self.steps.append([None, None])
		self.set_action(-1,action)
		return self

	def propagate(self, start_at=0):
		for i in range(start_at, len(self.steps) - 1):
			exp, act = self.steps[i]
			next_exp = self.steps[i+1][0]
			if exp != None and act != None and next_exp == None:
				self.set_expression(i+1, act.get_output_expression(exp))
		exp, act = self.steps[-1]
		if exp != None and act != None:
			try:
				self.add_expression_to_end(act.get_output_expression(exp))
			except NotImplementedError:
				pass

	def get_addressmap(self, index):
		exp, act = self.steps[index]
		return act.get_addressmap(exp)

	def get_animation(self, index, **kwargs):
		action = self.get_action(index)
		expA = self.get_expression(index)
		expB = self.get_expression(index+1)
		if action:
			Animation = action.get_animation()(expA, expB, **kwargs)
		else:
			Animation = TransformMatchingTex(expA.mob, expB.mob, **kwargs)
		return Animation
	
	def next_animation(self, **kwargs):
		return self.get_animation(self.current_exp_index, **kwargs)

	def play_animation(self, scene, index, **kwargs):
		expA = self.get_expression(index)
		expB = self.get_expression(index+1)
		if self.show_past_steps:
			self.shift_past_steps(scene, expA, expB)
		animation = self.get_animation(index, **kwargs)
		scene.play(animation)
		self.current_exp_index = index+1

	def play_next(self, scene):
		self.play_animation(scene, index=self.current_exp_index)

	def play_range(self, scene, start_index, end_index, wait_between=1, **kwargs):
		for i in range(start_index, end_index):
			self.play_animation(scene, index=i, **kwargs)
			scene.wait(wait_between)

	def play_all(self, scene, wait_between=0.5, reset_exp_index=False):
		if reset_exp_index:
			self.current_exp_index = 0
		if self.mob not in scene.mobjects:
			scene.play(Write(self.mob))
		while self.current_exp_index < len(self.steps)-1:
			scene.wait(wait_between)
			self.play_next(scene=scene)

	def shift_past_steps(self, scene, expA, expB):
		mobA_radius = expA.mob.get_critical_point(self.past_steps_direction) - expA.mob.get_center()
		mobB_radius = expB.mob.get_center() - expB.mob.get_critical_point(-self.past_steps_direction)
		shift_distance = np.linalg.norm(mobA_radius) + np.linalg.norm(mobB_radius) + self.past_steps_buff
		self.past_steps_vgroup.add(
			self.mob.copy().set_opacity(0.25)
		)
		scene.add(self.past_steps_vgroup)
		scene.play(
			self.past_steps_vgroup.animate.shift(shift_distance * self.past_steps_direction),
			run_time = self.past_steps_shift_run_time,
			rate_func = self.past_steps_shift_rate_func
		)

	def __repr__(self):
		return f"Timeline({str([[repr(exp), repr(act)] for exp, act in self.steps])})"

	def get_vgroup(self, **kwargs):
		return VGroup(*[self.steps[i][0].mob for i in range(len(self.steps))])
	
	@property
	def vgroup(self):
		return self.get_vgroup()

	@property
	def mob(self):
		return self.exp.mob

	@property
	def exp(self):
		return self.expressions[self.current_exp_index]

	@property
	def expressions(self):
		return [exp for exp,act in self.steps]

	@property
	def actions(self):
		return [act for exp,act in self.steps]

	def reset(self, reset_caches=True):
		self.current_exp_index = 0
		if reset_caches:
			for exp in self.expressions:
				exp.reset_caches()

	def undo_last_action(self):
		if self.actions[-1] is not None:
			self.steps[-1] = [self.steps[-1][0], None]
		else:
			self.steps.pop()
			self.undo_last_action()

	def reset_caches(self):
		for i,exp in enumerate(self.expressions):
			exp.reset_caches()
			self.set_expression(i, exp)

	def align_on_equals(self, strength=1):
		self.get_vgroup()
		equals_positions = [exp['='].get_center() for exp in self.expressions]
		avg = np.mean(np.stack(equals_positions), axis=0)
		for exp, pos in zip(self.expressions, equals_positions):
			exp.mob.shift(strength*(avg - pos))
		return self

	def get_mob_ladder(self):
		from MF_Tools.dual_compatibility import VGroup, ArcBetweenPoints, RIGHT, Text, ORANGE
		ladder = VGroup()
		mobs = self.get_vgroup().copy()
		ladder.expressions = mobs.arrange(DOWN, buff=1)
		ladder.arrows = VGroup(*[
			ArcBetweenPoints(
				np.array([mobs.get_edge_center(RIGHT)[0], m1.get_center()[1]-0.1, 0]),
				np.array([mobs.get_edge_center(RIGHT)[0], m2.get_center()[1]+0.1, 0]),
				angle=-3/4*PI
			).shift(0.75*RIGHT).set_stroke(width=2, opacity=0.5)
			for m1, m2 in zip(mobs[:-1], mobs[1:])
		])
		ladder.actions = VGroup(*[
			Text(repr(act)).scale(0.6).next_to(arrow, RIGHT, buff=0.25)
			for act, arrow in zip(self.actions, ladder.arrows)
		])
		ladder.addressmaps = VGroup(*[
			VGroup(*[
				Text(str(entry)).set_color(ORANGE)
				for entry in addressmap
			]).arrange(DOWN).scale(0.25).next_to(ladder.actions[i], DOWN)
			for i, addressmap in enumerate([
				act.get_addressmap(exp)
				for exp, act in zip(self.expressions[:-1], self.actions)
			])
		])
		ladder.add(ladder.expressions, ladder.arrows, ladder.actions, ladder.addressmaps)
		return ladder
	
	def save_to_file(self, filename):
		from ..utils import save_to_file
		for exp in self.expressions:
			exp._mob = None
		save_to_file(self, filename)

	def debug_anim(self, scene, i):
		scene.remove(self.mob)
		exp, act = self.steps[i]
		print('Input expression:', exp)
		print('Output expression:', act.get_output_expression(exp))
		print('Addressmap:')
		for entry in act.get_addressmap(exp):
			print(entry)
		self.play_animation(scene, i)

	def __le__(self, expr):
		assert isinstance(expr, Expression), "Can only apply expression >= timeline"
		timeline = self.copy()
		timeline >> expr
		return timeline.expressions[-1]



class TimelineScene(Scene):
	def __init__(self, *args, **kwargs):
		self.timeline = Timeline()
		super().__init__(*args, **kwargs)

	def add_and_play(self, action):
		self.timeline >> action
		self.timeline.play_all(self)

	def __and__(self, action):
		self.add_and_play(action)
		return self

	def suspend(self):
		self.timeline.suspend()

	def resume(self):
		self.timeline.resume()
	
	def add_ladder(self):
		self.ladder = self.timeline.get_mob_ladder()
		self.ladder.scale_to_fit(4, 8).center()
		self.clear()








from ..expressions.functions.functions import *
from .unit_circle import unit_circle_dict
import numpy as np



class TrigFunction(Function):
	def __init__(self,
		symbol = None,
		python_rule = None,
		**kwargs
	):
		super().__init__(
			symbol = symbol,
			python_rule = python_rule,
			parentheses_mode = 'weak',
			**kwargs
		)

	# def __pow__(self, other):
	# 	other = Smarten(other)
	# 	if other.compute() == -1:
	# 		return self.inverse()
	# 	else:
	# 		from .powers import TrigPower
	# 		return TrigPower(self, other)

	def inverse(self):
		pass
		from .inverses import Inverse



	


class Sine(TrigFunction):
	def __init__(self, **kwargs):
		super().__init__(
			symbol = '\\sin',
			symbol_glyph_length = 3,
			python_rule = np.sin,
			**kwargs
		)

	def evaluate(self, arg):
		if arg in unit_circle_dict:
			return unit_circle_dict[arg][0]
		else:
			return super().evaluate(arg)


class Cosine(TrigFunction):
	def __init__(self, **kwargs):
		super().__init__(
			symbol = '\\cos',
			symbol_glyph_length = 3,
			python_rule = np.cos,
			**kwargs
		)

	def evaluate(self, arg):
		if arg in unit_circle_dict:
			return unit_circle_dict[arg][1]
		else:
			return super().evaluate(arg)


class Tangent(TrigFunction):
	def __init__(self, **kwargs):
		super().__init__(
			symbol = '\\tan',
			symbol_glyph_length = 3,
			python_rule = np.tan,
			**kwargs
		)

	def evaluate(self, arg):
		if arg in unit_circle_dict:
			return unit_circle_dict[arg][2]
		else:
			return super().evaluate(arg)


class Cosecant(TrigFunction):
	def __init__(self, **kwargs):
		super().__init__(
			symbol = '\\csc',
			symbol_glyph_length = 3,
			python_rule = lambda x: 1/np.sin(x),
			**kwargs
		)

	def evaluate(self, arg):
		if arg in unit_circle_dict:
			return unit_circle_dict[arg][0].reciprocal()
		else:
			return super().evaluate(arg)


class Secant(TrigFunction):
	def __init__(self, **kwargs):
		super().__init__(
			symbol = '\\sec',
			symbol_glyph_length = 3,
			python_rule = lambda x: 1/np.cos(x),
			**kwargs
		)

	def evaluate(self, arg):
		if arg in unit_circle_dict:
			return unit_circle_dict[arg][1].reciprocal()
		else:
			return super().evaluate(arg)


class Cotangent(TrigFunction):
	def __init__(self, **kwargs):
		super().__init__(
			symbol = '\\cot',
			symbol_glyph_length = 3,
			python_rule = lambda x: 1/np.tan(x),
			**kwargs
		)

	def evaluate(self, arg):
		if arg in unit_circle_dict:
			return unit_circle_dict[arg][2].reciprocal()
		else:
			return super().evaluate(arg)


sin = Sine()
cos = Cosine()
tan = Tangent()
csc = Cosecant()
sec = Secant()
cot = Cotangent()




from ..expression_core import *
from .combiners import Combiner


class Operation:
	eval_op = None


class BinaryOperation(Combiner, Operation):
	def compute(self):
		result = self.children[0].compute()
		for child in self.children[1:]:
			result = self.eval_op(result, child.compute())
		return result

	def is_negative(self):
		return self.children[0].is_negative()


class Add(BinaryOperation):
	symbol = '+'
	symbol_glyph_length = 1
	eval_op = staticmethod(lambda x, y: x + y)


class Sub(BinaryOperation):
	symbol = '-'
	symbol_glyph_length = 1
	eval_op = staticmethod(lambda x, y: x - y)

	def auto_parentheses(self):
		for i, child in enumerate(self.children):
			if i > 0 and (isinstance(child, (Add, Sub)) or child.is_negative()):
				child.give_parentheses()
			child.auto_parentheses()
		return self


class Mul(BinaryOperation):
	eval_op = staticmethod(lambda x, y: x * y)
	def __init__(self, *args, mode='config', **kwargs):
		self._mode = mode
		super().__init__(*args, **kwargs)

	@property
	def mode(self):
		if self._mode == 'config':
			mode = algebra_config['multiplication_mode']
		else:
			mode = self._mode
		if mode == 'auto':
			mode = self.auto_determine_mode()
		return mode

	@property
	def symbol(self):
		symbol_dict = {
			'dot': '\\cdot',
			'x': '\\times',
			'juxtapose': ''
		}
		if self.mode in symbol_dict:
			return symbol_dict[self.mode]
		else:
			raise ValueError(f"Invalid multiplication mode: {self.mode}. Mode must be among {list(symbol_dict.keys())}")

	@property
	def symbol_glyph_length(self):
		glyph_length_dict = {
			'dot': 1,
			'x': 1,
			'juxtapose': 0
		}
		if self.mode in glyph_length_dict:
			return glyph_length_dict[self.mode]
		else:
			raise ValueError(f"Invalid multiplication mode: {self.mode}. Mode must be among {list(glyph_length_dict.keys())}")

	def auto_determine_mode(self):
		from ..numbers.number import Number
		if all(isinstance(child, Number) for child in self.children):
			return 'dot'
		else:
			return 'juxtapose'

	def auto_parentheses(self): # should be more intelligent based on mode
		for child in self.children:
			if isinstance(child, (Add, Sub)): # or child.is_negative():
				child.give_parentheses()
			child.auto_parentheses()
		return self


class Div(BinaryOperation):
	eval_op = staticmethod(lambda x, y: x / y)
	def __init__(self, *args, mode='fraction', **kwargs):
		self.mode = mode
		super().__init__(*args, **kwargs)

	@property
	def symbol(self):
		symbol_dict = {
			'fraction': '\\over',
			'inline': '\\div',
		}
		if self.mode in symbol_dict:
			return symbol_dict[self.mode]
		else:
			raise ValueError(f"Invalid division mode: {self.mode}. Mode must be among {list(symbol_dict.keys())}")

	@property
	def symbol_glyph_length(self):
		glyph_length_dict = {
			'fraction': 1,
			'inline': 1,
		}
		if self.mode in glyph_length_dict:
			return glyph_length_dict[self.mode]
		else:
			raise ValueError(f"Invalid division mode: {self.mode}. Mode must be among {list(glyph_length_dict.keys())}")

	def auto_parentheses(self):
		for child in self.children:
			if (isinstance(child, (Add, Sub, Mul, Div)) or child.is_negative()) and self.mode == 'inline':
				child.give_parentheses()
			child.auto_parentheses()
		return self

	def is_negative(self):
		return self.children[0].is_negative() or self.children[1].is_negative()
	
	def compute(self):
		num = self.children[0].compute()
		den = self.children[1].compute()
		if den == 0:
			raise ZeroDivisionError
		if num % den == 0:
			return int(num / den)
		else:
			return float(num) / float(den)


class Pow(BinaryOperation):
	symbol = '^'
	symbol_glyph_length = 0
	eval_op = staticmethod(lambda x, y: x ** y)

	def auto_parentheses(self):
		assert len(self.children) == 2, f'Children: {self.children}' #idc how to auto paren power towers
		if isinstance(self.children[0], BinaryOperation) or self.children[0].is_negative():
			self.children[0].give_parentheses()
		for child in self.children:
			child.auto_parentheses()
		return self

	def is_negative(self):
		return False


class UnaryOperation(Expression, Operation):
	symbol = None
	symbol_glyph_length = None
	eval_op = None

	@Expression.parenthesize_glyph_count
	def get_glyph_count(self):
		return self.symbol_glyph_length + self.children[0].glyph_count

	@Expression.parenthesize_latex
	def __str__(self):
		return self.symbol + '{' + str(self.children[0]) + '}'
	
	def compute(self):
		return self.eval_op(self.children[0].compute())

	special_character_to_glyph_method_dict = {
		**Expression.special_character_to_glyph_method_dict,
		'-': 'get_unary_glyph',
		'~': 'get_unary_glyph'
	}

	def get_unary_glyph(self):
		return list(range(0, self.symbol_glyph_length))
	
	@Expression.parenthesize_glyph_list
	def get_glyphs_at_addigit(self, addigit):
		if addigit == 0:
			start = 0
			start += self.symbol_glyph_length
			end = start + self.children[0].glyph_count
			return list(range(start, end))
		else:
			raise NotImplementedError(f"{self} has no children at index {addigit}")


class Negative(UnaryOperation):
	symbol = '-'
	symbol_glyph_length = 1
	eval_op = staticmethod(lambda x: -x)

	def auto_parentheses(self):
		if isinstance(self.children[0], (Add, Sub)) or self.children[0].is_negative():
			self.children[0].give_parentheses()
		self.children[0].auto_parentheses()
		return self

	def is_negative(self):
		return True


class PlusMinus(UnaryOperation):
	symbol = '\\pm'
	symbol_glyph_length = 1
	eval_op = None

	def auto_parentheses(self):
		if isinstance(self.children[0], (Add, Sub)) or self.children[0].is_negative():
			self.children[0].give_parentheses()
		self.children[0].auto_parentheses()
		return self


class MinusPlus(UnaryOperation):
	symbol = '\\mp'
	symbol_glyph_length = 1
	eval_op = None

	def auto_parentheses(self):
		if isinstance(self.children[0], (Add, Sub)) or self.children[0].is_negative():
			self.children[0].give_parentheses()
		self.children[0].auto_parentheses()
		return self


from .timeline_core import *
from .timeline_variants import *
from ..expressions.variables import Variable
from ..actions.evaluation import evaluate_
from ..expressions.combiners.relations import Relation


class Evaluate(AutoTimeline):
	first_expression = None
	mode = 'one at a time'
	allowed_type = Real
	def __init__(self,
		first_expression = None,
		mode = None,
		allowed_type = None,
		**kwargs
	):
		self.first_expression = self.first_expression or first_expression
		self.mode = mode or self.mode
		self.allowed_type = self.allowed_type or allowed_type
		super().__init__(**kwargs)
		if self.first_expression is not None:
			self.add_expression_to_start(self.first_expression)

	def decide_next_action(self, index: int, mode=None):
		if mode is None:
			mode = self.mode
		if mode == 'one at a time':
			last_exp = self.get_expression(index)
			twig_ads = last_exp.get_all_twig_addresses()
			for twig_ad in twig_ads:
				if not isinstance(last_exp.get_subex(twig_ad), Relation):
					try:
						action = evaluate_(allowed_type=self.allowed_type).pread(twig_ad)
						action.get_output_expression(last_exp)
						return action
					except (ValueError, IncompatibleExpression):
						pass
			return None
		elif mode == 'all at once':
			last_exp = self.get_expression(index)
			twig_ads = last_exp.get_all_twig_addresses()
			acceptable_twig_ads = []
			for twig_ad in twig_ads:
				try:
					action = evaluate_(allowed_type=self.allowed_type).pread(twig_ad)
					action.get_output_expression(last_exp)
					acceptable_twig_ads.append(twig_ad)
				except (ValueError, IncompatibleExpression):
					pass
			if len(acceptable_twig_ads) == 0:
				return None
			return evaluate_(allowed_type=self.allowed_type).pread(*acceptable_twig_ads)
		else:
			raise ValueError(f'Invalid mode: {mode}')



class Solve(AutoTimeline):
	def __init__(self, solve_for=None, first_expression=None, preferred_side=None, auto_evaluate=True, **kwargs):
		super().__init__(**kwargs)
		self.solve_for = solve_for
		self.auto_evaluate = auto_evaluate
		self.all_actions_to_try = []
		from ..algebra.equations import EquationManeuver
		for maneuver_ in EquationManeuver.__subclasses__():
			self.all_actions_to_try += [
				maneuver_(),
				maneuver_().reverse(),
				maneuver_().flip(),
				maneuver_().reverse_flip()
			]
		if first_expression is not None:
			self.add_expression_to_start(first_expression)

	def decide_next_action(self, index:int):
		last_exp = self.get_expression(index)
		if self.solve_for is None:
			self.solve_for = last_exp.get_all_variables().pop()
		current_addresses = last_exp.get_addresses_of_subex(self.solve_for)
		if len(current_addresses) != 1:
			print(f"Solve: I don't know what to do if variable appears {len(current_addresses)} times")
			return None
		current_address = current_addresses[0]

		if self.auto_evaluate:
			try:
				result = (Evaluate(last_exp)).decide_next_action(0)
				if result is not None:
					return result
			except IncompatibleExpression:
				pass

		if len(current_address) == 1:
			self._solved = True
			if current_address == '1':
				# return swap_children_()
				return None
			elif current_address == '0':
				return None
			else:
				raise Exception('What?', current_address)

		if self.solve_for is not None:
			successful_outputs = []
			for maneuver in self.all_actions_to_try:
				try:
					next_exp = maneuver.get_output_expression(last_exp)
					new_address = next_exp.get_addresses_of_subex(self.solve_for)[0]
					successful_outputs.append((maneuver, new_address))
				except IncompatibleExpression:
					pass
			
			if len(successful_outputs) == 0:
				return None
			shortest_result = min(successful_outputs, key=lambda p: len(p[1]))
			assert len(shortest_result[1]) <= len(current_address)
			return shortest_result[0]

		return None

	def set_solve_for(self, var):
		self.solve_for = var
		self.resume()
		return self
	
	@property
	def solution(self):
		last_exp = self.expressions[-1]
		solve_for_ad = last_exp.get_addresses_of_subex(self.solve_for)[0]
		if self._solved and len(solve_for_ad) == 1:
			return last_exp.get_subex(str(1-int(solve_for_ad)))
		else:
			return None
			

class EvaluateAndSolve(CombinedRuleTimeline):
	def __init__(self, *args, **kwargs):
		super().__init__(Evaluate, Solve, *args, **kwargs)


class SolveAndEvaluate(Solve, Evaluate):
	def __init__(self, *args, **kwargs):
		Solve.__init__(self, *args, **kwargs)
		Evaluate.__init__(self, *args, **kwargs)
	
	def decide_next_action(self, index):
		return super().decide_next_action(index)



from ..expression_core import *
from .number import *


class Integer(Number):
	value_type = int
	value_to_string_func = lambda n: str(n)
	# value_to_string_func = lambda n: '\\roman{' + str(n) + '}'
	# this allows global switch between roman, sumerian, etc numerals!
	def __init__(self, n, **kwargs):
		assert isinstance(n, int)
		super().__init__(**kwargs)
		self.value = n

	@Expression.parenthesize_glyph_count
	def get_glyph_count(self):
		return len(str(self.value))

	@Expression.parenthesize_latex
	def __str__(self):
		return Integer.value_to_string_func(self.value)
	
	def compute(self):
		return self.value

	def is_negative(self):
		return self.value < 0

	@staticmethod
	def GCF(*smartnums):
		smartnums = list(map(Smarten, smartnums))
		nums = list(map(lambda N: N.value, smartnums))
		return Smarten(int(np.gcd.reduce(nums)))

	@staticmethod
	def LCM(*smartnums):
		smartnums = list(map(Smarten, smartnums))
		nums = list(map(lambda N: N.value, smartnums))
		return Smarten(int(np.lcm.reduce(nums)))

	def prime_factorization(self):
		pass


zero = Integer(0)
one = Integer(1)
two = Integer(2)
three = Integer(3)
four = Integer(4)
five = Integer(5)
six = Integer(6)
seven = Integer(7)
eight = Integer(8)
nine = Integer(9)
ten = Integer(10)



from ..actions.action_core import Action, IncompatibleExpression
from ..expressions.variables import Variable
from ..expressions.functions import Function
from ..utils import Smarten
from MF_Tools.dual_compatibility import Write, FadeIn, FadeOut
from copy import deepcopy
from itertools import product


class AlgebraicAction(Action):
	template1 = None
	template2 = None
	addressmap = []
	auto_addressmap = False
	var_condition_dict = {}       # Example {c: lambda exp: isinstance(exp, Number)}
	var_kwarg_dict = {}           # Example {a:{'path_arc':PI}}

	def __init__(self,
		template1=None,
		template2=None,
		*extra_addressmaps,
		var_condition_dict={},
		var_kwarg_dict={},
		auto_addressmap=True,
		**kwargs
	):
		super().__init__(**kwargs)
		self.template1 = Smarten(self.template1) or Smarten(template1)
		self.template2 = Smarten(self.template2) or Smarten(template2)
		self.addressmap = self.addressmap or list(extra_addressmaps)
		self.var_condition_dict = self.var_condition_dict or var_condition_dict
		self.var_kwarg_dict = self.var_kwarg_dict or var_kwarg_dict
		self.auto_addressmap = self.auto_addressmap or auto_addressmap

	def get_output_expression(self, input_expression=None):
		var_dict = match_expressions(self.template1, input_expression, self.var_condition_dict)
		return self.template2.substitute(var_dict)

	def get_addressmap(self, input_expression=None):
		addressmap = [] if self.addressmap is None else list(self.addressmap)

		if not self.auto_addressmap:
			return addressmap

		def get_var_ad_dict(template):
			return {var: template.get_addresses_of_subex(var) for var in template.get_all_variables()}
		self.template1_address_dict = get_var_ad_dict(self.template1)
		self.template2_address_dict = get_var_ad_dict(self.template2)

		leaves = self.get_all_leaves()
		for leaf in leaves:
			kwargs = self.var_kwarg_dict.get(leaf, {})
			template1_addresses = self.template1_address_dict.get(leaf, [])
			template2_addresses = self.template2_address_dict.get(leaf, [])

			# print('leaves: ', leaves)
			# print('leaf: ', leaf)
			# print('template1_addresses: ', template1_addresses)
			# print('template2_addresses: ', template2_addresses)

			for t1ad, t2ad in product(template1_addresses, template2_addresses):
				if not any((t1ad, t2ad) == (str(ad[0]).strip('(_)'), str(ad[1]).strip('(_)')) for ad in addressmap):
					addressmap += [[t1ad, t2ad, kwargs]]
		
		return addressmap

	def __repr__(self):
		return f'AlgebraicAction({self.template1}, {self.template2})'
	
	def get_animation(self, *args, **kwargs):
		return super().get_animation(*args, auto_fade=True, auto_resolve_delay=0.1, **kwargs)

	def reverse(self):
		# swaps input and output templates
		result = self.copy()
		result.template1, result.template2 = result.template2, result.template1
		if result.addressmap is None:
			return result
		new_addressmap = deepcopy(result.addressmap)
		for entry in new_addressmap:
			entry[0], entry[1] = entry[1], entry[0]
			if len(entry) == 3 and 'path_arc' in entry[2].keys():
				entry[2]['path_arc'] = -entry[2]['path_arc']
			if entry[0] is FadeOut:
				entry[0] = self.introducer
			if entry[1] in (FadeIn, Write):
				entry[1] = self.remover
		result.addressmap = new_addressmap
		return result

	def get_all_variables(self):
		return self.template1.get_all_variables() | self.template2.get_all_variables()

	def get_all_leaves(self):
		return self.template1.get_all_leaves() | self.template2.get_all_leaves()

	def get_equation(self):
		assert self.template1 and self.template2
		from ..expressions.combiners.relations import Equation
		return Equation(self.template1, self.template2)



def match_expressions(template, expression, condition_dict={}):
	"""
		This function will either return a ValueError if the expression
		simply does not match the structure of the template, such as a missing
		operand or a plus in place of a times, or if they do match it will return
		a dictionary of who's who. For example,
		
		template:      (a*b)**n
		expression:    (4*x)**(3+y)
		return value:  {a:4, b:x, n:3+y}

		template:      n + x**5
		expression:    12 + x**3
		return value:  ValueError("Structures do not match at address 11, 5 vs 3")
		
		template:      x**n*x**m
		expression:    2**2*3**3
		return value:  ValueError("Conflicting matches for x: 2 and 3")

		Obviously this has to be recursive, but gee I am feeling a bit challenged atm...
		...
		Ok I think I've done it!
	"""
	# Leaf case
	if not template.children:
		condition = condition_dict.get(template, lambda exp: True)
		if not condition(expression):
			raise IncompatibleExpression(f"The subexpression {expression} is trying to match with {template}, but does not meet its given condition")
		if isinstance(template, Variable):
			return {template: expression}
		if isinstance(template, Function) and expression.is_function():
			return {template: expression}
		if template == expression:
			return {}
		raise IncompatibleExpression("Expressions do not match")
	
	# Node case
	var_dict = {}
	if not isinstance(expression, type(template)):
		raise IncompatibleExpression(f"Expressions do not match type: {expression} vs {template}")
	if not len(template.children) == len(expression.children):
		raise IncompatibleExpression("Expressions do not match children length")
	for tc,ec in zip(template.children, expression.children):
		child_dict = match_expressions(tc, ec, condition_dict)
		matching_keys = child_dict.keys() & var_dict.keys()
		if any(not child_dict[key] == var_dict[key] for key in matching_keys):
			raise IncompatibleExpression("Conflicting matches for " + str(matching_keys))
		var_dict.update(child_dict)

	return var_dict






from ..expressions.functions.functions import Function, c0, c1, arg
from ..expressions.combiners.relations import Equation


class IntegralOperator(Function):
	pass


class DefiniteIntegral(IntegralOperator):
	string_code = [
		lambda self: self.symbol,
		lambda self: '\\limits' if self.vertical_bounds else '',
		'_', c0, '^', c1, arg
		]
	glyph_code = [c1, 1, c0, arg]
	def __init__(self,
		start,
		end,
		variable = None,
		show_variable = False,
		vertical_bounds = False,
		**kwargs
		):
		self.variable = variable
		self.vertical_bounds = vertical_bounds
		if show_variable:
			assert variable is not None, "variable must be provided if show_variable is True"
			lower_bound = Equation(variable, start)
			upper_bound = Equation(variable, end)
		else:
			lower_bound = start
			upper_bound = end
		super().__init__(
			symbol = '\\int',
			symbol_glyph_length = 1,
			children = [lower_bound, upper_bound],
			parentheses_mode = 'weak',
			**kwargs
			)

	@property
	def lower_bound(self):
		return self.children[0]

	@property
	def upper_bound(self):
		return self.children[1]



class IndefiniteIntegral(IntegralOperator):
	def __init__(self, **kwargs):
		super().__init__(
			symbol = '\\int',
			symbol_glyph_length = 1,
			parentheses_mode = 'weak',
			**kwargs
			)

Integral = DefiniteIntegral
I = IndefiniteIntegral()



class PlugInBounds(Function):
	# string_code = ['\\quad', '\\left.', arg, '\\right\\rvert', '_', c0, '^', c1]
	string_code = ['\\quad', arg, '\\Bigg\\rvert', '_', c0, '^', c1]
	glyph_code = [arg, 3, c1, c0]
	def __init__(self, lower_bound, upper_bound, variable, show_variable=False):
		self.variable = variable
		self.show_variable = show_variable
		if show_variable:
			lower_bound = Equation(variable, lower_bound)
			upper_bound = Equation(variable, upper_bound)
		else:
			lower_bound = lower_bound
			upper_bound = upper_bound
		super().__init__(
			children = [lower_bound, upper_bound],
			parentheses_mode = 'weak'
		)
	
	@property
	def lower_bound(self):
		if self.show_variable:
			return self.get_subex('01')
		else:
			return self.get_subex('0')

	@property
	def upper_bound(self):
		if self.show_variable:
			return self.get_subex('11')
		else:
			return self.get_subex('1')

	def expand_on_args(self, arg):
		return arg @ {self.variable:self.upper_bound} - arg @ {self.variable:self.lower_bound}


from MF_Tools.dual_compatibility import (
	Text, dc_Tex as Tex,
	UP, DOWN, LEFT, RIGHT,
	GREEN, BLUE, ORANGE,
	Indicate,
	VGroup, VDict,
	Line,
	Scene
)
import numpy as np
from abc import ABC
from dataclasses import dataclass
from copy import deepcopy
from tabulate import tabulate



@dataclass
class MF_Base(ABC):
	def copy(self):
		return deepcopy(self)

	def __rshift__(self, other):
		other = Smarten(other)
		return combine_to_timeline(self, other)

	def __rrshift__(self, other):
		other = Smarten(other)
		return other.__rshift__(self)

	def __eq__(self, other):
		other = Smarten(other)
		if type(self) != type(other):
			return False
		else:
			return self.hash_key() == other.hash_key()

	def __hash__(self):
		return hash(self.hash_key())

	def hash_key(self):
		raise NotImplementedError


def Smarten(input):
	if input is None:
		return None

	if isinstance(input, MF_Base):
		return input.copy()

	if isinstance(input, int):
		from .expressions.numbers.integer import Integer
		return Integer(input)

	if isinstance(input, float):
		if input == np.inf:
			from .calculus.limits import inf
			return inf
		if input == np.nan:
			return None
		from math import isclose
		if isclose(input, round(input)):
			from .expressions.numbers.integer import Integer
			return Integer(round(input))
		from .expressions.numbers.real import Real
		return Real(input)

	if isinstance(input, complex):
		from .expressions.numbers.complex import Complex
		return Complex(input)

	if isinstance(input, tuple):
		from .expressions.combiners.sequences import Coordinate
		return Coordinate(*input)

	if input is ...:
		from .expressions.variables import dots
		return dots

	from decimal import Decimal
	if isinstance(input, Decimal):
		return Smarten(float(input))

	from fractions import Fraction
	if isinstance(input, Fraction):
		from .expressions.numbers.rational import Rational
		return Rational(input.numerator, input.denominator)

	raise NotImplementedError(f"Unsupported type {type(input)}")


def combine_to_timeline(A, B):
	from .expressions.expression_core import Expression
	from .actions.action_core import Action
	from .timelines.timeline_core import Timeline
	type_combo_to_timeline_func = {
		# (TypeofA, TypeofB) : function of A,B which returns the desired Timeline object
		(Expression, Expression) : lambda A,B: Timeline().add_expression_to_start(A).add_expression_to_end(B),
		(Expression, Action) : lambda A,B: Timeline().add_expression_to_start(A).add_action_to_end(B),
		(Action, Expression) : lambda A,B: Timeline().add_action_to_start(A).add_expression_to_end(B),
		(Action, Action) : lambda A,B: Timeline().add_action_to_start(A).add_action_to_end(B),
		(Expression, Timeline) : lambda A,B: B.add_expression_to_start(A),
		(Action, Timeline) : lambda A,B: B.add_action_to_start(A),
		(Timeline, Expression) : lambda A,B: A.add_expression_to_end(B),
		(Timeline, Action) : lambda A,B: A.add_action_to_end(B),
		(Timeline, Timeline) : lambda A,B: A.combine_timelines(B)
	}
	for (type1, type2), func in type_combo_to_timeline_func.items():
		if isinstance(A, type1) and isinstance(B, type2):
			return func(A, B)
	raise NotImplementedError(f"Unsupported combination of types {type(A)} and {type(B)}")


def add_spaces_around_brackets(input_string):
	result = []
	i = 0
	length = len(input_string)

	for i in range(length):
		if input_string[i:i+2] == '{{':
			result.append('{ ')
		elif input_string[i:i+2] == '}}':
			result.append('} ')
		else:
			result.append(input_string[i])

	# Join the list into a single string and remove any extra spaces
	return ''.join(result).strip()


def print_info(expression, tablefmt='rst'):
	def tree_prefix(address):
		V, T = '│ ', '├─'
		d = len(address)
		return (V * (d - 1) + T) if d else ''

	def get_all_info(expression, address):
		def get_info(callable):
			try:
				result = callable(expression, address)
				string = str(result)
			except Exception as e:
				string = str(e)
			max_length = 20
			if len(string) > max_length:
				string = string[:max_length-3] + '...'
			return string

		return {
			'Type': get_info(lambda Exp, ad: f'{tree_prefix(ad)}{type(Exp.get_subex(ad)).__name__}'),
			'Address': get_info(lambda Exp, ad: ad),
			'LaTeX string': get_info(lambda Exp, ad: str(Exp.get_subex(ad))),
			'glyph_count': get_info(lambda Exp, ad: str(Exp.get_subex(ad).glyph_count)),
			'glyph_indices': get_info(lambda Exp, ad: Exp.get_glyphs_at_address(ad)),
			'paren': get_info(lambda Exp, ad: Exp.get_subex(ad).parentheses),
			'color': get_info(lambda Exp, ad: getattr(Exp.get_subex(ad), 'color', None))
		}

	addresses = expression.get_all_addresses()
	rows = [get_all_info(expression, address) for address in addresses]
	table = tabulate(
		rows,
		headers='keys',
		tablefmt=tablefmt
	)
	print(table)


def random_number_expression(leaves=range(-5, 10), max_depth=3, max_children_per_node=2, **kwargs):
	import random
	from .expressions.numbers import Integer
	from .expressions.combiners.operations import Add, Sub, Mul, Div, Pow, Negative
	nodes = [Add, Sub, Mul, Pow]
	node = random.choice(nodes)
	def generate_child(current_depth):
		if np.random.random() < 1 / (current_depth + 1):
			return Integer(random.choice(leaves))
		else:
			return random_number_expression(leaves, max_depth - 1)
	def generate_children(current_depth, number_of_children):
		return [generate_child(current_depth) for _ in range(number_of_children)]
	if node == Add or node == Mul:
		children = generate_children(max_depth, random.choice(list(range(2,max_children_per_node+1))))
	elif node == Negative:
		children = generate_children(max_depth, 1)
	else:
		children = generate_children(max_depth, 2)
	return node(*children, **kwargs)


def create_graph(expr, node_size=0.5, horizontal_buff=1, vertical_buff=1.5, printing=False):
	def create_node(address):
		from .expressions.numbers import Integer, Real, Rational
		from .expressions.variables import Variable
		from .expressions.combiners.operations import Add, Sub, Mul, Div, Pow, Negative
		from .expressions.functions.functions import Function
		from .expressions.combiners.sequences import Sequence
		from .expressions.combiners.relations import Equation, LessThan, LessThanOrEqualTo, GreaterThan, GreaterThanOrEqualTo
		type_to_symbol_dict = {
			Integer: lambda expr: str(expr.value),
			Real: lambda expr: expr.symbol if expr.symbol else str(expr),
			Rational: lambda expr: '\\div',
			Variable: lambda expr: expr.symbol,
			Add: lambda expr: '+',
			Sub: lambda expr: '-',
			Mul: lambda expr: '\\times',
			Div: lambda expr: '\\div',
			Pow: lambda expr: '\\hat{}',
			Negative: lambda expr: '-',
			Function: lambda expr: expr.symbol,
			Sequence: lambda expr: ',',
			Equation: lambda expr: '=',
			LessThan: lambda expr: '<',
			LessThanOrEqualTo: lambda expr: '\\leq',
			GreaterThan: lambda expr: '>',
			GreaterThanOrEqualTo: lambda expr: '\\geq',
		}
		subex = expr.get_subex(address)
		symbol = type_to_symbol_dict[type(subex)](subex)
		tex = Tex(symbol)
		# if tex.width > tex.height:
		# 	tex.scale_to_fit_width(node_size)
		# else:
		# 	tex.scale_to_fit_height(node_size)
		return tex
	addresses = expr.get_all_addresses()
	if printing: print(addresses)
	max_length = max(len(address) for address in addresses)
	layered_addresses = [
		[ad for ad in addresses if len(ad) == i]
		for i in range(max_length + 1)
	]
	if printing: print(layered_addresses)
	max_index = max(range(len(layered_addresses)), key=lambda i: len(layered_addresses[i]))
	max_layer = layered_addresses[max_index]
	max_width = len(max_layer)
	if printing: print(max_index, max_width, max_layer)
	Nodes = VDict({ad: create_node(ad) for ad in addresses})
	#Max_layer = VGroup(*[Nodes[ad] for ad in max_layer]).arrange(RIGHT,buff=horizontal_buff)
	def position_children(parent_address):
		parent = Nodes[parent_address]
		child_addresses = [ad for ad in layered_addresses[len(parent_address)+1] if ad[:-1] == parent_address]
		if printing: print(child_addresses)
		child_Nodes = VGroup(*[Nodes[ad] for ad in child_addresses]).arrange(RIGHT,buff=1)
		child_Nodes.move_to(parent.get_center()+DOWN*vertical_buff)
	for i in range(max_index, max_length):
		for ad in layered_addresses[i]:
			position_children(ad)
	def position_parent(child_address):
		sibling_Nodes = VGroup(*[Nodes[ad] for ad in layered_addresses[len(child_address)] if ad[:-1] == child_address[:-1]])
		parent_Node = Nodes[child_address[:-1]]
		parent_Node.move_to(sibling_Nodes.get_center()+UP*vertical_buff)
	for i in range(max_index, 0, -1):
		for ad in layered_addresses[i]:
			position_parent(ad)
	Edges = VGroup(*[
		Line(
			Nodes[ad[:-1]].get_bounding_box_point(DOWN),
			Nodes[ad].get_bounding_box_point(UP),
			buff=0.2, stroke_opacity=0.4
			)
		for ad in addresses if len(ad) > 0
		])
	return VGroup(Nodes, Edges)


def to_sympy(exp):
	from sympy.parsing.latex import parse_latex
	latex = str(exp)
	sympy_expr = parse_latex(latex)

	# Special case substitution needed so that e is interpreted as the constant and not a variable
	from sympy import E, symbols
	sympy_expr = sympy_expr.subs(symbols('e'), E)

	return sympy_expr




def text_to_MF_Algebra(text):
	import re
	# Token regex
	token_re = re.compile(
		r"""
		(?<!\w)-?(?:\d+\.\d*|\.\d+|\d+)   # numbers
		| [a-zA-Z_]\w*                     # identifiers
		| \^                                # caret
		| =                                 # equals
		| [()+\-*/]                         # operators and parentheses
		""",
		re.VERBOSE
	)

	# Helper to categorize tokens
	def categorize(tok):
		if tok in '+-*/()|**':
			return 'op'
		if tok == '**':
			return 'caret'
		if tok == '|':
			return 'equals'
		if re.match(r'(?<!\w)-?(?:\d+\.\d*|\.\d+|\d+)$', tok):
			return 'number'
		return 'identifier'

	def rewrite_expression(s):
		# Step 1: tokenize & replace numbers/operators
		raw_tokens = token_re.findall(s)
		tokens = []
		for tok in raw_tokens:
			if tok == '^':
				tokens.append('**')
			elif tok == '=':
				tokens.append('|')
			elif re.match(r'(?<!\w)-?(?:\d+\.\d*|\.\d+|\d+)$', tok):
				tokens.append(f"Real({tok})" if '.' in tok else f"Integer({tok})")
			else:
				tokens.append(tok)

		# Step 2: insert implicit multiplication
		result = [tokens[0]]
		for prev, curr in zip(tokens, tokens[1:]):
			prev_cat = categorize(prev)
			curr_cat = categorize(curr)
			
			# new rules for implicit multiplication
			if ((prev_cat in ('number', 'identifier') or prev == ')') and
				(curr_cat in ('number', 'identifier') or curr == '(')):
				result.append('*')
			result.append(curr)
		return ''.join(result)

	from asteval import Interpreter
	from MF_Algebra import (Integer, Real, Variable, Function, sqrt, cbrt)
	symtable = {
		**{L: Variable(L, 1) for L in 'abcjklmnopqrstuvwxyz'},
		**{F: Function(F, 1) for F in 'fgh'},
		'Integer' : Integer,
		'Real' : Real,
		'Variable' : Variable,
		'sqrt' : sqrt,
		'cbrt' : cbrt,
	}
	aeval = Interpreter(symtable=symtable)

	return aeval(rewrite_expression(text))





import dill as pickle
import os

def save_to_file(obj:MF_Base, filename:str) -> None:
	os.makedirs('saved_objects', exist_ok=True)
	path = os.path.join('saved_objects', filename)
	with open(path, 'wb') as f:
		pickle.dump(obj, f)

def load_from_file(filename:str) -> MF_Base:
	path = os.path.join('saved_objects', filename)
	with open(path, 'rb') as f:
		return pickle.load(f)




def apply_addressmap(address:str, addressmap:list, reverse:bool=False) -> set:
	"""    
	Turns one address into another (or several), according to an addressmap.
	If it is explicitly listed in the addressmap, it simply returns the address on the other side.
	If it is not listed, it checks all the entries with that address as a prefix, and outputs the corresponding prefix on the other side.
	It also checks if any entry is a prefix of the address, in which cases it replaces that section of the address with the corresponding prefix.
	It may be found multiple times or not at all, which is why it returns a set. I think that makes sense but I'm not sure.
	The output set can have any number, you usually are expecting a singleton, empty or multiple usually means failure.
	Reverse simply goes from right to left across the addressmap instead of left to right.
	Returns None if address is ever mapped to [], meaning that subex is meant to fade in or out.
	
	Examples:
	>>> apply_addressmap('12', [['123', '999']])
	{'99'}
	>>> apply_addressmap('123', [['12', '88'], ['1', '5']])
	{'883', '523'}
	>>> apply_addressmap('12', [['1', []]])
	None
	"""

	la = len(address)
	from_ads = [entry[int(reverse)] for entry in addressmap]
	to_ads = [entry[int(not reverse)] for entry in addressmap]
	results = set()

	for from_ad, to_ad in zip(from_ads, to_ads):
		if isinstance(from_ad, type) or isinstance(to_ad, type):
			# Address is being explicitly animated in or out
			return None

		lf,lt = len(from_ad), len(to_ad)

		# Case 1: address is equal to from_ad or some prefix of it
		if address == from_ad[:la]:
			if to_ad == []: return None
			results.add(to_ad[:la-lf or None])

		# Case 2: from_ad is a prefix of address
		elif from_ad == address[:lf]:
			if to_ad == []: return None
			results.add(to_ad + address[lf:])

	return results


from ..expressions.functions import ApplyFunction
from ..expressions.functions.operators import Sum
from ..expressions.variables import n,k,x
from .limits import inf


class Series(ApplyFunction):
	def __init__(self,
		general_term,
		variable = n,
		start = 0,
		end = inf,
		**kwargs
	):
		variables = general_term.get_all_variables()
		if variable not in variables and len(variables) > 0:
			variable = variables.pop()
		self.variable = variable
		self.start = start
		self.end = end
		self.term = general_term
		self.sigma = Sum(variable, start, end)
		super().__init__(self.sigma, self.term, **kwargs)






from .set_core import Set
from ..expressions import Expression, Sequence


class Interval(Set, Sequence):
	def __init__(self, a, b, **kwargs):
		super().__init__(a, b, **kwargs)
		assert a < b, "Interval lower bound must be less than upper bound"
		self.a = a
		self.b = b
		
	def in_rule(self, x):
		if isinstance(x, Expression):
			try: x = x.compute()
			except: return False
		if x < self.a: return False
		elif x == self.a: return self.paren_symbols[0] == '['
		elif self.a < x < self.b: return True
		elif x == self.b: return self.paren_symbols[1] == ']'
		elif self.b < x: return False



class OO_Interval(Interval):
	parentheses = True
	paren_symbols = ('(',')')

class CC_Interval(Interval):
	parentheses = True
	paren_symbols = ('[',']')

class OC_Interval(Interval):
	parentheses = True
	paren_symbols = ('(',']')

class CO_Interval(Interval):
	parentheses = True
	paren_symbols = ('[',')')




from ..expressions.functions.functions import Function, arg
from ..expressions.variables import x, y, z, r, theta, s, t, u, v


class DifferentialOperator(Function):
	pass


d = DifferentialOperator(
	symbol = '\\mathrm{d}',
	symbol_glyph_length = 1,
	parentheses_mode = 'weak'
)

dx = d(x)
dy = d(y)
dz = d(z)
dr = d(r)
dtheta = d(theta)
ds = d(s)
dt = d(t)
du = d(u)
dv = d(v)


def dd(var):
    return d/d(var)


class PrimeDiff(DifferentialOperator):
	string_code = [arg, '^\\prime']
	glyph_code = [arg, 1]




from .expression_core import *


class Variable(Expression):
	def __init__(self, symbol, symbol_glyph_length=None, **kwargs):
		self.symbol = symbol
		self.symbol_glyph_length = symbol_glyph_length
		super().__init__(**kwargs)

	@Expression.parenthesize_glyph_count
	def get_glyph_count(self):
		if algebra_config['fast_glyph_count'] and not self.symbol_glyph_length:
			raise ValueError(f'Fast glyph count mode is on but {self} has no set glyph_length')
		return self.symbol_glyph_length

	@Expression.parenthesize_latex
	def __str__(self):
		return self.symbol

	def compute(self):
		raise ValueError(f"Expression contains a variable {self.symbol}.")

	def hash_key(self):
		return (self.__class__, tuple(self.children), self.symbol)



a = Variable('a', 1)
b = Variable('b', 1)
c = Variable('c', 1)
# d = Variable('d', 1) # Differential
# e = Variable('e', 1) # Real
# f = Variable('f', 1) # Function
# g = Variable('g', 1) # Function
# h = Variable('h', 1) # Function
# i = Variable('i', 1) # Complex
j = Variable('j', 1)
k = Variable('k', 1)
l = Variable('l', 1)
m = Variable('m', 1)
n = Variable('n', 1)
o = Variable('o', 1)
p = Variable('p', 1)
q = Variable('q', 1)
r = Variable('r', 1)
s = Variable('s', 1)
t = Variable('t', 1)
u = Variable('u', 1)
v = Variable('v', 1)
w = Variable('w', 1)
x = Variable('x', 1)
y = Variable('y', 1)
z = Variable('z', 1)

alpha = Variable('\\alpha', 1)
beta = Variable('\\beta', 1)
gamma = Variable('\\gamma', 1)
theta = Variable('\\theta', 1)
phi = Variable('\\phi', 1)

dots = Variable('\\ldots', 3)
cdots = Variable('\\cdots', 3)


class Variables(ExpressionContainer):
	expression_type = Variable




from .action_core import Action


class ParallelAction(Action):
    def __init__(self, *actions, lag=0, **kwargs):
        self.actions = list(actions)
        self.lag = lag
        super().__init__(**kwargs)
    
    def get_output_expression(self, input_expression=None):
        expr = input_expression
        for action in self.actions:
            expr = action.get_output_expression(expr)
        return expr

    def get_addressmap(self, input_expression=None):
        total_addressmap = []
        for i,act in enumerate(self.actions):
            for entry in act.get_addressmap(input_expression):
                if len(entry) == 2: entry.append({})
                entry[2]['delay'] = entry[2].get('delay', 0) + i*self.lag
                total_addressmap.append(entry)
        return total_addressmap

    def __or__(self, other):
        if isinstance(other, ParallelAction):
            return ParallelAction(*self.actions, *other.actions)
        elif isinstance(other, Action):
            return ParallelAction(*self.actions, other)
        else:
            raise ValueError("Can only use | with other ParallelAction or Action")
    
    def __ror__(self, other):
        if isinstance(other, Action):
            return ParallelAction(other, *self.actions)
        else:
            return NotImplemented





from .action_core import Action
from MF_Tools.dual_compatibility import PI


class Permutation:
	def __init__(self, *order):
		# Permutation(1,0,3,2)
		self.order = order

	def __len__(self):
		return len(self.order)

	def __call__(self, n):
		assert isinstance(n, int) and n >= 0
		if 0 <= n < len(self.order):
			return self.order[n]
		else:
			return n

	def __matmul__(self, other):
		assert isinstance(other, Permutation)
		N = max(len(self), len(other))
		return Permutation(*[self(other(n)) for n in range(N)])

	def inverse_call(self, n):
		if n in self.order:
			return self.order.index(n)
		else:
			return n

	def __invert__(self):
		return Permutation(*[self.inverse_call(n) for n in range(len(self))])

	@classmethod
	def from_cycles(cls, *cycles):
		order = []
		all_values_present = {n for cycle in cycles for n in cycle}
		max_val = max(all_values_present)
		for i in range(max_val+1):
			if i in all_values_present:
				for cycle in cycles:
					if i in cycle:
						index = cycle.index(i)
						index += 1
						value = cycle[index % len(cycle)]
						order.append(value)
						break
			else:
				order.append(i)
			i += 1
		return cls(*order)

	def __repr__(self):
		return 'Permutation' + str(self.order)



class permute_children_(Action):
	def __init__(self, permutation, mode='arc', arc_size=0.75*PI, **kwargs):
		self.permutation = permutation
		self.mode = mode
		self.arc_size = arc_size
		super().__init__(**kwargs)

	def get_output_expression(self, input_expression=None):
		input_expression.children = [
			input_expression.children[self.permutation(i)]
			for i in range(len(input_expression.children))
		]
		return input_expression

	def get_addressmap(self, input_expression=None):
		assert len(input_expression.children) >= len(self.permutation)
		
		if self.mode == 'arc':
			kwarg_dict = {'path_arc': self.arc_size}
		elif self.mode == 'straight':
			kwarg_dict == {}
		else:
			raise ValueError(f'Invalid mode: {self.mode}. Must be \'arc\' or \'straight\'.')
		
		addressmap = [
			[str(n), str(self.permutation(n)), kwarg_dict]
			for n in range(len(self.permutation))
		]

		from ..expressions.combiners import Combiner
		if isinstance(input_expression, Combiner):
			addressmap.append(['+', '+', kwarg_dict])

		return addressmap


class swap_children_(permute_children_):
	def __init__(self, **kwargs):
		super().__init__(
			permutation = Permutation(1,0),
			**kwargs
		)



from ..expressions import Expression


class Set(Expression):
	in_rule = None
	_set = set()

	def __init__(self,
		*children,
		in_rule = None, # condition to check if input is in set
		**kwargs
	):
		super().__init__(*children, **kwargs)
		self.in_rule = self.in_rule or in_rule

	def __contains__(self, item):
		# x in A (must return bool)
		if self.in_rule:
			return self.in_rule(Smarten(item))
		else:
			return item in self._set

	def __iter__(self):
		return iter(self._set)

	def __and__(self, other):
		from .set_operations import Intersection
		return Intersection(self, other)

	def __rand__(self, other):
		from .set_operations import Intersection
		return Intersection(other, self)

	def __or__(self, other):
		from .set_operations import Union
		return Union(self, other)
	
	def __ror__(self, other):
		from .set_operations import Union
		return Union(other, self)

	def __sub__(self, other):
		from .set_operations import Difference
		return Difference(self, other)
	
	def __rsub__(self, other):
		from .set_operations import Difference
		return Difference(other, self)

	def __truediv__(self, other):
		from .set_operations import Difference
		return Difference(self, other)
	
	def __rtruediv__(self, other):
		from .set_operations import Difference
		return Difference(other, self)
	
	def __rmod__(self, other):
		from .set_operations import IsElement
		return IsElement(other, self)
	
	def __lt__(self, other):
		from .set_operations import IsProperSubset
		return IsProperSubset(self, other)
	
	def __le__(self, other):
		from .set_operations import IsSubset
		return IsSubset(self, other)
	
	def __mul__(self, other):
		from .set_operations import CartesianProduct
		return CartesianProduct(self, other)
	
	def __rmul__(self, other):
		from .set_operations import CartesianProduct
		return CartesianProduct(other, self)

	def __pow__(self, other):
		from .set_operations import CartesianPower
		return CartesianPower(self, other)
	
	def __rpow__(self, other):
		from .set_operations import CartesianPower
		return CartesianPower(other, self)



from ..expressions import Sequence
class ElementsSet(Set, Sequence):
	parentheses = True
	paren_symbols = ('\\{','\\}')

	def __init__(self,
		*children,
		in_rule = None, #condition to check if input is in set
		**kwargs
	):
		super().__init__(*children, **kwargs)
		self._set = set(children)
		self.in_rule = in_rule or self.in_rule


from ..expressions import Variable
class SymbolSet(Set, Variable):
	symbol = None
	in_rule = None


from ..expressions import ExpressionContainer
class Sets(ExpressionContainer):
	expression_type = SymbolSet


from ..expressions import Combiner
class SetBuilder(Set, Combiner):
	symbol = '|'
	symbol_glyph_length = 1
	left_spacing = '\\enspace'
	right_spacing = '\\enspace'
	parentheses = True
	paren_symbols = ('\\{','\\}')

	def __init__(self, expr, *conditions, **kwargs):
		super().__init__(expr, Sequence(*conditions), **kwargs)
	










from ..expressions import Integer, Real, dots
from ..utils import Smarten

Empty = SymbolSet(symbol='\\varnothing', symbol_glyph_length=1, in_rule=lambda x: False)
Empty.elements = ElementsSet()

Z = SymbolSet(
	symbol = '\\mathbb{Z}',
	symbol_glyph_length = 1,
	in_rule = lambda x: isinstance(Smarten(x).compute(), Integer)
	)
Z.elements = ElementsSet(dots, -2, -1, 0, 1, 2, dots)

N = SymbolSet(
	symbol = '\\mathbb{N}',
	symbol_glyph_length = 1,
	in_rule = lambda x: isinstance(Smarten(x).compute(), Integer) and x.value >= 0
)
N.elements = ElementsSet(0, 1, 2, 3, dots)

R = SymbolSet(
	symbol = '\\mathbb{R}', 
	symbol_glyph_length = 1,
	in_rule = lambda x: isinstance(Smarten(x).compute(), (Integer, Real))
)







from .action_core import Action, IncompatibleExpression
from ..expressions.combiners.operations import Operation
from ..expressions.numbers.number import Number


class evaluate_(Action):
	def __init__(self, mode='random leaf', allowed_type=Number, **kwargs):
		self.mode = mode # Idk if we will use this, seems like more of a Timeline decision
		self.allowed_type = Number
		super().__init__(**kwargs)

	def get_output_expression(self, input_expression=None):
		try:
			output = input_expression.evaluate()
			assert isinstance(output, self.allowed_type)
			return output
		except:
			raise IncompatibleExpression

	def get_addressmap(self, input_expression=None):
		return [
			['', ''] # Extension by preaddress is done by decorator!
		]
	



from ..actions.action_core import Action
from ..expressions.combiners.operations import Add, Sub, Mul, Div, Pow


class distribute_(Action):
	# Not done yet, multilayer does not work... this is so necessary but rather nontrivial... hm...
	def __init__(self, mode='auto', multilayer=False, **kwargs):
		self.mode = mode #'auto', 'left', 'right'
		super().__init__(**kwargs)

	def get_output_expression(self, input_expression=None):
		if self.mode == 'auto':
			self.determine_direction(input_expression)
		if self.mode == 'left':
			new_children = [
				type(input_expression)(input_expression.children[0], child)
				for child in input_expression.children[-1].children
				]
			return type(input_expression.children[-1])(*new_children)
		elif self.mode == 'right':
			new_children = [
				type(input_expression)(child, input_expression.children[-1])
				for child in input_expression.children[0].children
				]
			return type(input_expression.children[0])(*new_children)

	def determine_direction(self, input_expression=None):
		if self.mode == 'auto':
			if isinstance(input_expression, Mul):
				left_distributable = isinstance(input_expression.children[-1], (Add, Sub))
				right_distributable = isinstance(input_expression.children[0], (Add, Sub))
				if left_distributable and right_distributable:
					raise ValueError('Cannot auto-distribute if both sides are distributable, please set mode manually.')
				elif left_distributable:
					self.mode = 'left'
				elif right_distributable:
					self.mode = 'right'
				else:
					raise ValueError('Cannot distribute, neither side is distributable.')
			elif isinstance(input_expression, Div):
				right_distributable = isinstance(input_expression.children[0], (Add, Sub))
				if right_distributable:
					self.mode = 'right'
				else:
					raise ValueError('Cannot distribute, right side is not distributable.')
			elif isinstance(input_expression, Pow):
				right_distributable = isinstance(input_expression.children[0], (Mul, Div))
				if right_distributable:
					self.mode = 'right'
				else:
					raise ValueError('Cannot distribute, right side is not distributable.')
			else:
				raise ValueError('Cannot auto-distribute, must be a multiplication or division.')

	def get_addressmap(self, input_expression=None):
		return [
			['', ''] #standin idk what the fuck im doing here
		]



from .combiners import Combiner


class Script(Combiner):
	def auto_parentheses(self):
		from .operations import Operation
		for i,child in enumerate(self.children):
			if i==0 and isinstance(child, (Combiner, Operation)):
				child.give_parentheses()
			child.auto_parentheses()
		return self

	def is_variable(self):
		from ..variables import Variable
		return isinstance(self.children[0], Variable)

class Subscript(Script):
	symbol = '_'
	symbol_glyph_length = 0

class Superscript(Script):
	symbol = '^'
	symbol_glyph_length = 0


from ..variables import x,y,z

x1 = Subscript(x,1)
x2 = Subscript(x,2)
x3 = Subscript(x,3)
y1 = Subscript(y,1)
y2 = Subscript(y,2)
y3 = Subscript(y,3)
z1 = Subscript(z,1)
z2 = Subscript(z,2)
z3 = Subscript(z,3)



from .functions import Function, c0, c1, arg
from ..combiners.relations import Equation
from ..combiners.operations import *


class BigOperator(Function):
	string_code = [lambda self: self.symbol, '\\limits', '_', c0, '^', c1, arg]
	glyph_code = [c1, 1, c0, arg]
	OpClass = None
	def __init__(self, variable, start, end, **kwargs):
		super().__init__(
			children=[Equation(variable, start), end],
			parentheses_mode='weak', 
			**kwargs
		)

	@property
	def variable(self):
		return self.get_subex('00')

	@property
	def start(self):
		return self.get_subex('01').compute()

	@property
	def end(self):
		return self.get_subex('1').compute()

	def expand_on_args(self, arg, min=None, max=None, max_num_terms=7, substitute=True):
		if min is None:
			min = self.start
		if max is None:
			max = self.end
		assert min <= max
		if max - min > max_num_terms:
			terms = [
				arg @ {self.variable : i} if substitute else arg
				for i in range(min, min+max_num_terms)
			]
			from ..variables import cdots
			terms.append(cdots)
		else:
			terms = [
				arg @ {self.variable : i} if substitute else arg
				for i in range(min, max+1)
			]
		return self.OpClass(*terms)


class Sum(BigOperator):
	OpClass = Add
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.symbol = '\\sum'
		self.symbol_glyph_length = 1


class Product(BigOperator):
	OpClass = Mul
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.symbol = '\\prod'
		self.symbol_glyph_length = 1









from .algebra_core import AlgebraicAction
from ..actions.permutations import swap_children_
from ..expressions.variables import a,b,c
from ..expressions.functions.radicals import Rad, sqrt
from ..expressions.functions.logarithms import Log, ln
from ..expressions.numbers.real import e
from ..expressions.combiners.operations import PlusMinus
from MF_Tools.dual_compatibility import FadeIn, FadeOut, Write, PI
from copy import deepcopy


class EquationManeuver(AlgebraicAction):
	# Watch out, I don't think these cannot be preaddressed currently.
	# But I can't conceive of why you'd want to do that anyway.
	# Perhaps for a sequence of equations?

	def reverse(self):
		# swaps input and output templates
		self.template1, self.template2 = self.template2, self.template1
		# modifies addressmap accordingly
		# (swap order and negate path_arcs)
		admap_copy = deepcopy(self.addressmap)
		for entry in admap_copy:
			entry[0], entry[1] = entry[1], entry[0]
			if len(entry) == 3 and 'path_arc' in entry[2].keys():
				entry[2]['path_arc'] = -entry[2]['path_arc']
		self.addressmap = admap_copy

		return self

	def flip(self):
		# flips the equations of both templates
		s = swap_children_()
		self.template1 = s.get_output_expression(self.template1)
		self.template2 = s.get_output_expression(self.template2)
		# modifies addressmap accordingly
		# (swap first character of addresses and negate path_arcs)
		admap_copy = deepcopy(self.addressmap)
		for entry in admap_copy:
			if isinstance(entry[0], str):
				entry[0] = str(1-int(entry[0][0])) + entry[0][1:]
			if isinstance(entry[1], str):
				entry[1] = str(1-int(entry[1][0])) + entry[1][1:]
			if len(entry) == 3 and 'path_arc' in entry[2].keys():
				entry[2]['path_arc'] = -entry[2]['path_arc']
		self.addressmap = admap_copy

		return self

	def reverse_flip(self):
		return self.reverse().flip()


class alg_add_R(EquationManeuver):
	template1 = a + b | c
	template2 = a | c - b
	addressmap = (
		['01', '11', {'path_arc':PI}],
		['0+', '1-', {'path_arc':PI}],
	)


class alg_add_L(EquationManeuver):
		template1 = a + b | c
		template2 = b | c - a
		addressmap = (
			['00', '11', {'path_arc':PI}],
			['0+', '1-', {'path_arc':PI}]
		)
		# return (
		#     ['00', '11', {'path_arc':PI}],
		#     ['0+', FadeOut, {'run_time':0.5}],
		#     [FadeIn, '1-', {'run_time':0.5, 'delay':0.5}]
		# )


class alg_mul_R(EquationManeuver):
	template1 =	a * b | c
	template2 = a | c / b
	addressmap = (
		['01', '11', {'path_arc':PI}],
		['0*', [], {'run_time':0.5}],
		[[], '1/', {'run_time':0.5, 'delay':0.5}]
	)
	
	def reverse(self):
		super().reverse()
		self.addressmap = (
			['11', '01', {'path_arc':-PI}],
			[Write, '0*', {'run_time':0.5, 'delay':0.5}],
			['1/', FadeOut, {'run_time':0.5}]
		)
		return self


class alg_mul_L(EquationManeuver):
	template1 = a * b | c
	template2 = b | c / a
	addressmap = (
		['00', '11', {'path_arc':PI}],
		['0*', [], {'run_time':0.5}],
		[[], '1/', {'run_time':0.5, 'delay':0.5}]
	)

	def reverse(self):
		super().reverse()
		self.addressmap = (
			['11', '00', {'path_arc':-PI}],
			[Write, '0*', {'run_time':0.5, 'delay':0.5}],
			['1/', FadeOut, {'run_time':0.5}]
		)
		return self


class alg_pow_2_R(EquationManeuver):
	template1 = a**2 | b
	template2 = a | PlusMinus(sqrt(b))
	addressmap = (
		['01', '10f', {'path_arc':-PI/3}],
		[[], '1~']
	)
	
	def reverse(self): # Hopefully this never happens lol
		super().reverse()
		self.addressmap = (
			['10f', '01', {'path_arc':PI/3}],
		)
		return self


class alg_pow_R(EquationManeuver):
	template1 = a**b | c
	template2 = a | Rad(b)(c)
	addressmap = (
		['01', '100', {'path_arc':-PI/3}],
		[[], '10f', {'delay':0.5, 'run_time':0.5}],
	)
	
	def reverse(self):
		super().reverse()
		self.addressmap = (
			['100', '01', {'path_arc':PI/3}],
			['10f', [], {'run_time':0.5}]
		)
		return self


class alg_pow_e_L(EquationManeuver):
	template1 = e**a | b
	template2 = a | ln(b)
	addressmap = (
		['00', '10f', {'path_arc':PI/2}],
	)
	
	def reverse(self):
		super().reverse()
		self.addressmap = (
			['10f', '00', {'path_arc':-PI/2}],
		)
		return self


class alg_pow_L(EquationManeuver):
	template1 = a**b | c
	template2 = b | Log(a)(c)
	addressmap = (
		['00', '100', {'path_arc':PI/2}],
		[[], '10f', {'delay':0.75, 'run_time':0.75}],
	)
	
	def reverse(self):
		super().reverse()
		self.addressmap = (
			['100', '00', {'path_arc':-PI/2}],
			['10f', FadeOut, {'run_time':0.5}]
		)
		return self


class alg_neg_R(EquationManeuver):
	template1 = -a | b
	template2 = a | -b
	addressmap = (
		['0-', '1-', {'path_arc':PI}],
	)



from .action_core import Action
from ..expressions.combiners.operations import Add, Sub, Mul, Div, Pow
from ..expressions.combiners.relations import Equation
from ..expressions.functions.functions import ApplyFunction
from ..utils import Smarten
from MF_Tools.dual_compatibility import Write


class apply_operation_(Action):
	OpClass = None
	side = 'right'
	introducer = Write

	def __init__(self, other, OpClass=None, side=None, introducer=None, **kwargs):
		self.other = Smarten(other)
		self.OpClass = OpClass or self.OpClass
		self.side = side or self.side
		self.introducer = introducer or self.introducer
		super().__init__(**kwargs)

	def get_output_expression(self, input_expression):
		if self.side == 'right':
			output_expression = self.OpClass(input_expression, self.other)
		elif self.side == 'left':
			output_expression = self.OpClass(self.other, input_expression)
		else:
			raise ValueError(f'Invalid side: {self.side}. Must be left or right.')
		return output_expression

	def get_addressmap(self, input_expression):
		if self.side == 'right':
			return [
				['', '0'],
				[[], '+', {'delay':0.5}],
				[[], '1', {'delay':0.6}]
			]
		elif self.side == 'left':
			return [
				['', '1'],
				[[], '0', {'delay':0.5}],
				[[], '+', {'delay':0.6}]
			]
		else:
			raise ValueError(f'Invalid side: {self.side}. Must be left or right.')

	def __repr__(self):
		return type(self).__name__ + '(' + str(self.other) + ')'


class add_(apply_operation_):
	OpClass = Add

class sub_(apply_operation_):
	OpClass = Sub

class mul_(apply_operation_):
	OpClass = Mul

class div_(apply_operation_):
	OpClass = Div

class pow_(apply_operation_):
	OpClass = Pow

class equals_(apply_operation_):
	OpClass = Equation

class apply_func_(apply_operation_):
	OpClass = ApplyFunction
	side = 'left'




from MF_Tools.dual_compatibility import dc_Tex, MANIM_TYPE, VGroup, WHITE, BLACK
from ..utils import MF_Base, Smarten, add_spaces_around_brackets
from functools import wraps


algebra_config = {
	'auto_parentheses': True,
	'multiplication_mode': 'auto',
	'division_mode': 'fraction',
	'decimal_precision': 4,
	'always_color': {},
	'default_color': WHITE,
	'fast_paren_length': True,
	'fast_glyph_count': True,
	'fast_root_length': True,
}


class Expression(MF_Base):
	parentheses = False
	paren_symbols = ('(', ')')
	def __init__(self, *children):
		self.children = list(map(Smarten,children))
		if algebra_config['auto_parentheses']:
			self.auto_parentheses()
		self.reset_caches()

	def reset_caches(self):
		self._mob = None
		self._glyph_count = None
		return self


	### Mobject ###

	@property
	def mob(self):
		if self._mob is None:
			self.init_mob()
		return self._mob

	def init_mob(self, **kwargs):
		string = add_spaces_around_brackets(str(self))
		self._mob = dc_Tex(string, **kwargs)
		self._mob.set_color(algebra_config['default_color'])
		self.set_color_by_subex(algebra_config['always_color'])

	def __getitem__(self, key):
		# Returns a VGroup of the glyphs at the given addresses
		# Or if key is an Expression, returns the glyphs of that subexpression!
		if MANIM_TYPE == 'GL':
			parent = self.mob
		elif MANIM_TYPE == 'CE':
			parent = self.mob[0]
		else:
			raise Exception(f"Unknown manim type: {MANIM_TYPE}")
		
		result = VGroup()
		if isinstance(key, int):
			# result.add(parent[key]) # Turning to Expression instead
			result.add(*self[Smarten(key)])
		elif isinstance(key, slice):
			result.add(*parent[key]) # Slice will still do mob I guess
		elif isinstance(key, str):
			result.add(*[parent[g] for g in self.get_glyphs_at_address(key)])
		elif isinstance(key, (list, tuple)):
			for k in key:
				result.add(*self[k])
		elif isinstance(key, Expression):
			for ad in self.get_addresses_of_subex(key):
				if not key.parentheses:
					ad += '_'
				result.add(*self[ad])
		else:
			raise ValueError(f"Invalid key: {key}")
		return result


	### Glyphs ###

	@property
	def glyph_count(self):
		# Set this value in subclasses so as not to need to render latex
		if self._glyph_count is None:
			self.init_glyph_count()
		return self._glyph_count

	def init_glyph_count(self):
		if algebra_config['fast_glyph_count']:
			try:
				gc = self.get_glyph_count()
				assert isinstance(gc, int)
				self._glyph_count = gc
				return
			except (NotImplementedError, AssertionError, AttributeError, ValueError):
				pass
		gc = self.get_glyph_count_from_mob()
		self._glyph_count = gc

	def get_glyph_count(self):
		# Guesses the number of glyphs in the expression, from a formula for the subclass.
		# Override in subclasses
		raise NotImplementedError

	def get_glyph_count_from_mob(self):	
		if MANIM_TYPE == 'GL':
			parent = self.mob
		elif MANIM_TYPE == 'CE':
			parent = self.mob[0]
		else:
			raise Exception(f"Unknown manim type: {MANIM_TYPE}")
		return len(parent)

	special_character_to_glyph_method_dict = {
		# Class dictionary mapping special characters to methods
		# When a character is seen in an address, the corresponding method is called
		# Subclasses can add entries to this dictionary
		'(': 'get_left_paren_glyphs',
		')': 'get_right_paren_glyphs',
		'_': 'get_exp_glyphs_without_parentheses',
		'#': 'get_glyphs_at_all_children', # This would be a cool idea, so 10#1 is equivalent to 1001,1011,1021 or whatever
	}

	def get_glyphs_at_address(self, address):
		# Returns the list of glyph indices at the given address
		if len(address) == 0:
			return list(range(self.glyph_count))

		addigit = address[0]
		remainder = address[1:]
		result = []

		if addigit == '!': # This can't be caught in the dictionary because it needs to use remainder
			return self.get_glyphs_except(remainder)

		if addigit in self.special_character_to_glyph_method_dict:
			glyph_method = getattr(self, self.special_character_to_glyph_method_dict[addigit])
			result += glyph_method()
			if remainder:
				result += self.get_glyphs_at_address(remainder)
			return sorted(list(set(result)))

		else:
			try:
				digit = int(addigit)
				child_glyphs = self.get_glyphs_at_addigit(digit)
				if len(child_glyphs) == 0:
					return []
				child = self.children[digit]
				glyphs_within_child = child.get_glyphs_at_address(remainder)
				shift_value = child_glyphs[0]
				result = [glyph + shift_value for glyph in glyphs_within_child]
				return sorted(list(set(result)))
			except:
				raise ValueError(f"Invalid address: {address}")

	def get_glyphs_at_addresses(self, *addresses):
		result = []
		for address in addresses:
			result += self.get_glyphs_at_address(address)
		return sorted(list(set(result)))

	def get_left_paren_glyphs(self):
		if not self.parentheses:
			return []
		start = 0
		end = start + self.paren_length()
		return list(range(start, end))

	def get_right_paren_glyphs(self):
		if not self.parentheses:
			return []
		end = self.glyph_count
		start = end - self.paren_length()
		return list(range(start, end))

	def get_exp_glyphs_without_parentheses(self):
		start = 0
		end = self.glyph_count
		if self.parentheses:
			start += self.paren_length()
			end -= self.paren_length()
		return list(range(start, end))

	def get_glyphs_except(self, address):
		return sorted(list(
			set(range(self.glyph_count)) -
			set(self.get_glyphs_at_address(address))
			))

	def __len__(self):
		return self.glyph_count

	def get_glyphs_at_addigit(self, addigit:int):
		raise NotImplementedError #Implement in subclasses


	### Addresses ###

	def get_all_addresses(self):
		# Returns the addresses of all subexpressions
		addresses = ['']
		for n in range(len(self.children)):
			for child_address in self.children[n].get_all_addresses():
				addresses.append(str(n)+child_address)
		return sorted(list(set(addresses)))

	def get_all_addresses_with_condition(self, condition):
		result = set()
		for address in self.get_all_addresses():
			if condition(self.get_subex(address)):
				result |= {address}
		return sorted(list(result))

	def get_all_leaf_addresses(self):
		return self.get_all_addresses_with_condition(
			lambda subex: not subex.children
		)

	def get_all_twig_addresses(self):
		return self.get_all_addresses_with_condition(
			lambda subex: subex.children and all(
				not child.children or child.is_function()
				for child in subex.children
			)
		)

	def get_all_addresses_of_type(self, expression_type):
		return self.get_all_addresses_with_condition(
			lambda subex: isinstance(subex, expression_type)
		)

	def get_addresses_of_subex(self, target_subex):
		return self.get_all_addresses_with_condition(
			lambda subex: subex == target_subex
		)

	def get_all_addresses_with_paren(self):
		return self.get_all_addresses_with_condition(
			lambda subex: subex.parentheses
		)


	### Subexpressions ###

	def get_subex(self, address_string):
		# Returns the Expression object corresponding to the subexpression at the given address.
		if address_string == '':
			return self
		elif int(address_string[0]) < len(self.children):
			return self.children[int(address_string[0])].get_subex(address_string[1:])
		else:
			raise IndexError(f"No subexpression of {self} at address {address_string} .")

	def get_subexes_at_addresses(self, *addresses):
		return {self.get_subex(ad) for ad in addresses}

	def get_all_subexpressions_with_condition(self, condition):
		result = set()
		for address in self.get_all_addresses():
			if condition(subex := self.get_subex(address)):
				result |= {subex}
		return result

	def get_all_subexpressions(self):
		return self.get_all_subexpressions_with_condition(lambda subex: True)

	def get_all_subexpressions_of_type(self, expression_type):
		return self.get_all_subexpressions_with_condition(lambda subex: isinstance(subex, expression_type))

	def get_all_numbers(self):
		from .numbers.number import Number
		return self.get_all_subexpressions_of_type(Number)

	def get_all_variables(self):
		# Doing it this way so as to catch subscripted variables and not their parent
		if self.is_variable():
			return {self}
		results = set()
		for child in self.children:
			if child.is_variable():
				results |= {child}
			else:
				results |= child.get_all_variables()
		return results

	def get_all_leaves(self):
		return self.get_subexes_at_addresses(*self.get_all_leaf_addresses())

	def get_all_subex_with_paren(self):
		return self.get_all_subexpressions_with_condition(lambda subex: subex.parentheses)

	@property
	def left(self):
		return self.children[0]

	@property
	def right(self):
		return self.children[1]


	### Operations ###

	def __neg__(self):
		from .combiners.operations import Negative
		return Negative(self)

	def __add__(self, other):
		from .combiners.operations import Add
		return Add(self, other)

	def __radd__(self, other):
		from .combiners.operations import Add
		return Add(other, self)

	def __sub__(self, other):
		from .combiners.operations import Sub
		return Sub(self, other)

	def __rsub__(self, other):
		from .combiners.operations import Sub
		return Sub(other, self)

	def __mul__(self, other):
		from .combiners.operations import Mul
		return Mul(self, other)

	def __rmul__(self, other):
		from .combiners.operations import Mul
		return Mul(other, self)

	def __truediv__(self, other):
		from .combiners.operations import Div
		return Div(self, other)

	def __rtruediv__(self, other):
		from .combiners.operations import Div
		return Div(other, self)

	def __pow__(self, other):
		from .combiners.operations import Pow
		return Pow(self, other)

	def __rpow__(self, other):
		from .combiners.operations import Pow
		return Pow(other, self)

	def __and__(self, other):
		from .combiners.relations import Equation
		return Equation(self, other)

	def __rand__(self, other):
		from .combiners.relations import Equation
		return Equation(other, self)

	def __or__(self, other):
		from .combiners.relations import Equation
		return Equation(self, other)

	def __ror__(self, other):
		from .combiners.relations import Equation
		return Equation(other, self)

	def __lt__(self, other):
		from .combiners.relations import LessThan
		return LessThan(self, other)

	def __le__(self, other):
		from .combiners.relations import LessThanOrEqualTo
		return LessThanOrEqualTo(self, other)

	def __gt__(self, other):
		from .combiners.relations import GreaterThan
		return GreaterThan(self, other)

	def __ge__(self, other):
		other = Smarten(other)
		if isinstance(other, Expression):
			from .combiners.relations import GreaterThanOrEqualTo
			return GreaterThanOrEqualTo(self, other)
		else:
			return NotImplemented

	def __matmul__(self, other):
		if isinstance(other, dict):
			return self.substitute(other)
		other = Smarten(other)
		if self.is_function() and other.is_function():
			from .functions.functions import Composition
			return Composition(self, other)
		else:
			return NotImplemented


	### Parentheses ###

	def give_parentheses(self, parentheses=True, symbols=None):
		change = parentheses - self.parentheses
		if change or symbols:
			self.reset_caches()
			self.parentheses = parentheses
			if symbols:
				self.paren_symbols = symbols
		return self

	def clear_all_parentheses(self):
		for child in self.children:
			child.clear_all_parentheses()
		self.give_parentheses(False)
		return self

	def auto_parentheses(self):
		for child in self.children:
			child.auto_parentheses()
		return self

	def reset_parentheses(self):
		self.clear_all_parentheses()
		self.auto_parentheses()
		return self

	def paren_length(self):
		# Returns the number of glyphs taken up by the expression's potential parentheses.
		# Usually 1 but can be larger for larger parentheses.
		if algebra_config['fast_paren_length'] == True:
			return 1
		yes_paren = self.copy().give_parentheses(True)
		no_paren = self.copy().give_parentheses(False)
		num_paren_glyphs = yes_paren.glyph_count - no_paren.glyph_count
		assert num_paren_glyphs > 0 and num_paren_glyphs % 2 == 0
		return num_paren_glyphs // 2

	@staticmethod
	def parenthesize_latex(str_func):
		# To decorate most subclasses' __str__ methods
		@wraps(str_func)
		def wrapper(expr, *args, **kwargs):
			pretex = str_func(expr, *args, **kwargs)
			if expr.parentheses:
				p1,p2 = expr.paren_symbols
				pretex = '\\left' + p1 + pretex + '\\right' + p2
			return pretex
		return wrapper

	@staticmethod
	def parenthesize_glyph_count(gc_func):
		# To decorate most subclasses' get_glyph_count methods
		@wraps(gc_func)
		def wrapper(expr, *args, **kwargs):
			gc = gc_func(expr, *args, **kwargs)
			if isinstance(gc, int) and expr.parentheses:
				gc += 2 * expr.paren_length()
			return gc
		return wrapper

	@staticmethod
	def parenthesize_glyph_list(gl_func):
		# Shifts lists from get_glyphs_at_addigit by paren length when necessary
		@wraps(gl_func)
		def wrapper(expr, *args, **kwargs):
			gl = gl_func(expr, *args, **kwargs)
			shift = expr.parentheses * expr.paren_length()
			if shift:
				gl = [g + shift for g in gl]
			return gl
		return wrapper


	### Substitution ###

	def substitute_at_address(self, subex, address):
		subex = Smarten(subex)
		if len(address) == 0:
			return subex
		index = int(address[0])
		result = self.copy()
		new_child = result.children[index].substitute_at_address(subex, address[1:])
		result.children[index] = new_child
		Expression.__init__(result, *result.children)
		return result

	def substitute_at_addresses(self, subex, addresses):
		result = self.copy()
		for address in addresses:
			result = result.substitute_at_address(subex, address)
		return result

	def substitute(self, expression_dict):
		result = self.copy()
		dict_with_numbers = list(enumerate(expression_dict.items()))
		from .variables import Variable
		for i, (from_subex, to_subex) in dict_with_numbers:
			result = result.substitute_at_addresses(Variable(f'T_{i}'), result.get_addresses_of_subex(from_subex))
		for i, (from_subex, to_subex) in dict_with_numbers:
			result = result.substitute_at_addresses(to_subex, result.get_addresses_of_subex(Variable(f'T_{i}')))
		return result


	### Functions ###

	def is_function(self):
		from .functions.functions import Function
		return isinstance(self, Function) or any(child.is_function() for child in self.children)

	def __call__(self, *inputs):
		assert self.is_function(), 'Tried to function call an expression that cannot be interpreted as a function'
		from .functions.functions import ApplyFunction
		if len(inputs) == 1:
			arg = Smarten(inputs[0])
		elif len(inputs) > 1:
			from .combiners.sequences import Sequence
			arg = Sequence(*list(map(Smarten, inputs)))
		return ApplyFunction(self, arg)


	### Nesting ###
	# These do not work yet, regrettably
	def nest(self, direction='right', recurse=True):
		if len(self.children) <= 2:
			return self
		else:
			if direction == 'right':
				return type(self)(self.children[0], type(self)(*self.children[1:]).nest(direction, recurse))
			elif direction == 'left':
				return type(self)(type(self)(*self.children[:-1]).nest(direction, recurse), self.children[-1])
			else:
				raise ValueError(f"Invalid direction: {direction}. Must be right or left.")

	def denest(self, denest_all = False, match_type = None):
		if len(self.children) <= 1:
			return self
		if match_type is None:
			match_type = type(self)
		new_children = []
		for child in self.children:
			if type(child) == match_type:
				for grandchild in child.children:
					new_children.append(grandchild.denest(denest_all, match_type))
			elif denest_all:
				new_children.append(child.denest(True, match_type))
			else:
				new_children.append(child)
		return type(self)(*new_children)


	### Coloring ###

	def set_color_by_subex(self, subex_color_dict):
		for subex, color in subex_color_dict.items():
			for ad in self.get_addresses_of_subex(subex):
				self.get_subex(ad).color = color
				if self.get_subex(ad).parentheses and not subex.parentheses:
					ad += '_'
				self[ad].set_color(color)
		return self

	def get_color_of_subex(self, subex): # This is awful lol
		for ad in self.get_addresses_of_subex(subex):
			subex = self.get_subex(ad)
			if hasattr(subex, 'color'):
				return subex.color		


	### Utilities ###

	def hash_key(self):
		return (self.__class__, tuple(self.children))

	def __repr__(self):
		max_length = 1e6
		string = type(self).__name__ + '(' + self.repr_string() + ')'
		if len(string) > max_length:
			string = string[:max_length-3] + '...'
		return string

	def repr_string(self):
		return self.__class__.__str__.__wrapped__(self) # Can be overriden in subclasses with an annoying latex string

	def is_negative(self):
		return False # catchall if not defined in subclasses

	def is_variable(self):
		from .variables import Variable
		return isinstance(self, Variable)

	def compute(self):
		# Define for operations, functions, etc
		raise NotImplementedError

	def evaluate(self):
		return Smarten(self.compute())

	def reciprocal(self):
		from .combiners.operations import Div
		if isinstance(self, Div):
			return Div(*self.children[::-1])
		else:
			return Div(1, self)

	def get_addressbook(self):
		all_addresses = self.get_all_addresses()
		for address in all_addresses.copy():
			if self.get_subex(address).parentheses:
				all_addresses.append(address + '_')
		return {
			address: self.get_glyphs_at_address(address)
			for address in all_addresses
		}

	def get_subexpression_chain(self, address):
		if address == '':
			return [self]
		else:
			return self.get_subexpression_chain(address[:-1]) + [self.get_subex(address)]



class Address:
	def __init__(self, *address_parts):
		if len(address_parts) == 1:
			arg = address_parts[0]
			if isinstance(arg, (str, list, tuple)):
				self.addigits = list(arg)
			elif isinstance(arg, int):
				self.addigits = [arg]
			elif isinstance(arg, Address):
				self.addigits = arg.addigits
			else:
				raise ValueError(f"Invalid address: {arg}")
		else:
			self.addigits = list(address_parts)

	def __getitem__(self, index):
		return self.addigits[index]

	def __len__(self):
		return len(self.addigits)

	def __repr__(self):
		return f'Address({self.addigits})'

	@classmethod
	def from_ints(cls, *ints):
		return cls(*[str(i) for i in ints])





class ExpressionContainer:
	expression_type = Expression
	# Nothing but a container so that you can make multiple expressions at a time.
	# Subclasses: Variables, Sets
	# a,b,c = Variables('abc')
	# mu,chi,psi = Variables('\\mu', '\\chi', '\\psi')
	# A,B,C = Sets('ABC')
	# You can set the glyph lengths in a list but it's really not a big deal if you don't.
	def __init__(self, *strings, symbol_glyph_length_list=None):
		if len(strings) == 1:
			self.strings = list(strings[0])
		else:
			self.strings = list(strings)
		self.symbol_glyph_length_list = symbol_glyph_length_list

	def __iter__(self):
		if self.symbol_glyph_length_list:
			assert len(self.strings) == len(self.symbol_glyph_length_list)
			return (
				self.expression_type(symbol=s, symbol_glyph_length=l)
				for s, l in zip(self.strings, self.symbol_glyph_length_list)
			)
		else:
			return (
				self.expression_type(symbol=s)
				for s in self.strings
				)



from ..expression_core import *
from .functions import Function, child, arg
from ..numbers.integer import two


class Rad(Function):
	string_code = ['\\sqrt', '[', child, ']', arg]
	glyph_code = [child, lambda self: self.radical_glyph_count(), arg]
	def __init__(self, index, **kwargs):
		super().__init__(
			symbol = '\\sqrt',
			children = [index],
			parentheses_mode = 'never',
			**kwargs
        )

	def python_rule(self, x):
		index = self.index.compute()
		return x**(1/index)

	@property
	def index(self):
		return self.children[0]

	def radical_glyph_count(self):
		if algebra_config['fast_root_length']:
			return 2
		else:
			raise NotImplementedError
	
	# def expand_on_args(self, arg, mode='rational_exponent'):
	# 	if mode == 'rational_exponent':
	# 		return arg ** (1 / self.index)
	# 	else:
	# 		raise NotImplementedError


class SquareRoot(Rad):
	string_code = ['\\sqrt', arg]
	glyph_code = [2, arg]
	standard_form = Rad(2)
sqrt = SquareRoot(2)


cbrt = Rad(3)



from ..expression_core import *
import numpy as np


class Number(Expression):
	value_type = None
	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.value = None

	def compute(self):
		return float(self)

	def __float__(self):
		return float(self.value)
	
	def __int__(self):
		return int(self.value)

	def hash_key(self):
		return (self.__class__, tuple(self.children), self.value)



from ..expression_core import *
from ..combiners.sequences import Sequence
from ..combiners.operations import BinaryOperation
from ..variables import Variable


arg = Variable('arg')
arg_num = lambda n: Variable(f'arg {n}')
arg0 = arg_num(0)
arg1 = arg_num(1)
arg2 = arg_num(2)

child_num = lambda n: Variable(f'child {n}')
child = child_num(0)
c0 = child_num(0)
c1 = child_num(1)
c2 = child_num(2)


# Example glyph and string codes

# Sigma
string_code = ['\\sum_', c0, '^', c1, arg]
glyph_code = [c1, 1, c0, arg]

# Nth Root
string_code = ['\\sqrt[', c0, ']', arg]
glyph_code = [c0, 2, arg]

# Default
string_code = [lambda self: self.symbol, arg]
glyph_code = [lambda self: self.symbol_glyph_length, arg]


class Function(Expression):
	string_code = [lambda self: self.symbol, arg]
	glyph_code = [lambda self: self.symbol_glyph_length, arg]
	def __init__(self,
		symbol = None,
		symbol_glyph_length = None,
		python_rule = None,
		algebra_rule_variables = None,
		algebra_rule = None,
		parentheses_mode = 'always',
		children = [],
		**kwargs
	):
		self.symbol = symbol #string
		self.symbol_glyph_length = symbol_glyph_length #int
		self._python_rule = python_rule #callable
		self.algebra_rule = algebra_rule
		if algebra_rule_variables is not None:
			self.algebra_rule_variables = algebra_rule_variables
		elif algebra_rule is not None:
			self.algebra_rule_variables = algebra_rule.get_all_variables()
		self.parentheses_mode = parentheses_mode
		super().__init__(*children, **kwargs)

	@property
	def python_rule(self):
		if self._python_rule:
			return self._python_rule
		else:
			raise NotImplementedError

	@Expression.parenthesize_latex
	def __str__(self):
		return self.get_string_from_code()

	def get_string_from_code(self):
		string = ''
		for sc in self.string_code:
			string += get_value_from_code_entry(self, sc, str)
		return string

	@Expression.parenthesize_glyph_count
	def get_glyph_count(self):
		count = 0
		for gc in self.glyph_code:
			count += get_value_from_code_entry(self, gc, int)
		return count

	@Expression.parenthesize_glyph_list
	def get_glyphs_at_addigit(self, addigit):
		start = 0
		for gc in self.glyph_code:
			if isinstance(gc, Variable) and gc == child_num(addigit):
				end = start + self.children[addigit].glyph_count
				return list(range(start, end))
			else:
				start += get_value_from_code_entry(self, gc, int)
		return []

	special_character_to_glyph_method_dict = {
		**Expression.special_character_to_glyph_method_dict,
		'f': 'get_main_func_glyphs',
		'F': 'get_all_func_glyphs',
		'c': 'get_all_child_glyphs',
	}

	def get_all_child_glyphs(self):
		glyphs = []
		for addigit in range(len(self.children)):
			glyphs += self.get_glyphs_at_addigit(addigit)
		return sorted(glyphs)

	def get_main_func_glyphs(self):
		return sorted(list(set(self.get_all_func_glyphs()) - set(self.get_all_child_glyphs())))

	def get_all_func_glyphs(self): # Needs glyph_code rework...
		# The trouble is we need the glyph counts for the argument(s) in case there are after them like n!
		return list(range(0, self.glyph_count))

	def is_function(self):
		return True

	def compute(self):
		raise ValueError('Functions should not be computed directly. It can be computed on its arguments by the ApplyFunction operation.')

	def compute_on_args(self, *computed_args):
		if self.python_rule is not None:
			return self.python_rule(*computed_args)
		elif self.algebra_rule and self.algebra_rule_variables:
			def rule(*args):
				assert len(args) == len(self.algebra_rule_variables), 'Mismatched number of arguments'
				substitution = {var: arg for var, arg in zip(self.algebra_rule_variables, args)}
				result = self.algebra_rule @ substitution
				return result.compute()
			self.python_rule = rule
			return rule(*computed_args)
		else:
			raise NotImplementedError('No python_rule or algebra_rule defined for this function')

	def expand_on_args(self, *arg_expressions):
		if self.algebra_rule is not None:
			assert len(self.algebra_rule_variables) == len(arg_expressions), 'Mismatched number of arguments'
			return self.algebra_rule @ {var: arg for var, arg in zip(self.algebra_rule_variables, arg_expressions)}
		else:
			raise NotImplementedError

	def evaluate(self, *arg_expressions):
		# Override for special values like trig
		computed_args = [arg.compute() for arg in arg_expressions]
		computed_value = self.compute_on_args(*computed_args)
		return Smarten(computed_value)

	def hash_key(self):
		return (self.__class__, tuple(self.children), self.symbol)

	@property
	def equation(self, **kwargs):
		from ..combiners.relations import Equation
		return Equation(self(*self.algebra_rule_variables), self.algebra_rule)

	@property
	def lambdify(self):
		assert self.algebra_rule is not None, 'This function has no algebra rule'
		from ...utils import to_sympy
		from sympy import lambdify
		sympy_vars = [to_sympy(v) for v in self.algebra_rule_variables]
		sympy_rule = to_sympy(self.algebra_rule)
		return lambdify(sympy_vars, sympy_rule, 'math')



class ApplyFunction(BinaryOperation):
	symbol = ''
	symbol_glyph_length = 0
	def __init__(self, *children, **kwargs):
		assert len(children) == 2
		assert children[0].is_function()
		super().__init__(*children, **kwargs)
		if not hasattr(self.func, 'string_code'):
			self.func.string_code = [lambda self: str(self), arg]
		if not hasattr(self.func, 'glyph_code'):
			self.func.glyph_code = [lambda self: self.glyph_count, arg]

	def compute(self):
		computed_args = [arg.compute() for arg in self.args_list]
		return self.func.compute_on_args(*computed_args)

	@property
	def func(self):
		return self.children[0]

	@property
	def arg(self):
		return self.children[1]

	@property
	def args_list(self):
		if isinstance(self.arg, Sequence):
			return [*self.arg.children]
		else:
			return [self.arg]

	@Expression.parenthesize_latex
	def __str__(self):
		return self.get_string_from_string_code()

	def get_string_from_string_code(self):
		string = ''
		for sc in self.func.string_code:
			string += get_value_from_code_entry(self, sc, str)
		return string

	@Expression.parenthesize_glyph_count
	def get_glyph_count(self):
		count = 0
		for gc in self.func.glyph_code:
			count += get_value_from_code_entry(self, gc, int)
		return count

	@Expression.parenthesize_glyph_list
	def get_glyphs_at_addigit(self, addigit):
		all_glyphs = set(range(0, self.glyph_count))
		arg_glyphs = set(self.get_arg_glyphs())
		if addigit == 0:
			return sorted(list(all_glyphs - arg_glyphs))
		if addigit == 1:
			return sorted(list(arg_glyphs))

	def get_arg_glyphs(self):
		start = 0
		start += self.parentheses * self.paren_length()
		for gc in self.func.glyph_code:
			if isinstance(gc, Variable) and gc.symbol == 'arg':
				end = start + self.arg.glyph_count
				return list(range(start, end))
			else:
				start += get_value_from_code_entry(self, gc, int)
		raise ValueError('arg not found in glyph_code')

	def auto_parentheses(self):
		from ..combiners.sequences import Sequence
		from ..combiners.operations import Add, Sub, Mul
		# Usually False, true for say (f+g)(x)
		if isinstance(self.func, (Sequence, Add, Sub, Mul, Composition)):
			self.func.give_parentheses(True)
		self.func.auto_parentheses()

		parentheses_mode = getattr(self.func, 'parentheses_mode', 'always')
		if parentheses_mode == 'always':
			self.arg.give_parentheses(True)
		from ..combiners.operations import Operation
		if parentheses_mode == 'strong' and (isinstance(self.arg, Operation) or self.arg.is_negative()):
			self.arg.give_parentheses(True)
		from ..combiners.operations import Add, Sub
		if parentheses_mode == 'weak' and (isinstance(self.arg, (Add, Sub)) or self.arg.is_negative()):
			self.arg.give_parentheses(True)
		if parentheses_mode == 'var only' and not self.arg.is_variable():
			self.arg.give_parentheses(True)
		if parentheses_mode == 'never':
			self.arg.give_parentheses(False)
		self.arg.auto_parentheses()

		return self

	def is_function(self):
		# This should always be False because it can no longer be called on something
		# Maybe not in very bold circumstances, but this will do for now
		return False

	def expand_on_args(self):
		return self.func.expand_on_args(*self.args_list)

	def evaluate(self):
		try:
			result = self.expand_on_args()
		except NotImplementedError:
			result = self.func.evaluate(*self.args_list)
		return result

	def equation_from_args(self, **kwargs):
		from ..combiners.relations import Equation
		return Equation(self, self.expand_on_args(**kwargs))


class Composition(BinaryOperation):
	symbol = '\\circ'
	symbol_glyph_length = 1
	eval_op = staticmethod(lambda x,y: Expression.__call__(x,y)) # ???
	string_code = [c0, lambda self: self.symbol, c1]
	glyph_code = [c0, lambda self: self.symbol_glyph_length, c1]
	def __init__(self, *children, **kwargs):
		assert len(children) == 2
		assert children[0].is_function() and children[1].is_function()
		super().__init__(*children, **kwargs)


def get_value_from_code_entry(expression, entry, desired_type):
	if isinstance(entry, desired_type):
		return entry

	if isinstance(expression, Function):
		func = expression
	elif isinstance(expression, ApplyFunction):
		func = expression.func
		arg = expression.arg
	else:
		raise ValueError('Function or ApplyFunction expected')

	if isinstance(entry, Variable):
		if entry.symbol.startswith('child'):
			func_child_number = int(entry.symbol.split()[-1])
			target = func.children[func_child_number]
		elif entry.symbol.startswith('arg'):
			if isinstance(expression, Function):
				return desired_type()
			if entry.symbol == 'arg':
				target = arg
			else:
				arg_child_number = int(entry.symbol.split()[-1])
				target = func.children[arg_child_number]
		else:
			raise ValueError('Invalid code variable: ' + entry.symbol)
		if desired_type == int:
			return target.glyph_count
		elif desired_type == str:
			return '{' + str(target) + '}'
		else:
			raise ValueError('Invalid desired_type. Must be int or str')

	if callable(entry):
		return desired_type(entry(func))

	raise ValueError(f'Invalid glyph code entry of type {type(entry)}: {entry}')



f = Function('f', 1)
g = Function('g', 1)
h = Function('h', 1)


from ..expression_core import ExpressionContainer
class Functions(ExpressionContainer):
	expression_type = Function



from .timeline_core import *


class AutoTimeline(Timeline):
	def __init__(self, auto_fill=True, **kwargs):
		super().__init__(**kwargs)
		self.auto_fill = auto_fill
	
	def set_expression(self, index: int, expression: Expression):
		super().set_expression(index, expression)
		if self.auto_fill and self.get_action(index) is None:
			next_action = self.decide_next_action(index)
			if next_action is not None:
				self.set_action(index, next_action)
		return self
	
	def set_action(self, index: int, action: Action):
		super().set_action(index, action)
		if self.auto_fill and self.get_expression(index+1) is None:
			next_expression = action.get_output_expression(self.get_expression(index))
			self.set_expression(index+1, next_expression)
		return self
	
	def decide_next_action(self, index: int) -> Action:
		# Implement in subclasses. Return None if finished.
		return None

	def suspend(self):
		self.auto_fill = False
		return self

	def resume(self):
		self.auto_fill = True
		self.add_action_to_end(self.decide_next_action(-1))
		return self


class CombinedRuleTimeline(AutoTimeline):
	def __init__(self, *timelines, **kwargs):
		self.constituents = [timeline if isinstance(timeline, type) else timeline.__class__ for timeline in timelines]
		# for timeline in self.constituents:
		#     timeline.__init__(self, **kwargs)
		super().__init__(**kwargs)
	
	def decide_next_action(self, index: int) -> Action:
		for timeline in self.constituents:
			result = timeline.decide_next_action(self, index)
			if result is not None:
				return result
		return None




